/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 * Portions Copyright 2013 Jean-Pierre Andre
 */

/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <sys/param.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/strsun.h>
#include <sys/policy.h>
#include <inet/common.h>
#include <inet/nd.h>
#include <inet/mi.h>
#include <sys/note.h>
#include <sys/mac_provider.h>
//#include "net80211_impl.h"

#include "queue.h"
#include "pe_var.h"
#include "resource_var.h"
#include "ntoskrnl_var.h"
#include "ndis_var.h"
#include "hal_var.h"
#include "cfg_var.h"
#include <if_ndisvar.h>
#include <inet/wifi_ioctl.h>

#define BRUSSELSSUPPORT 1

static int wl_set_essid(struct ndis_softc*, const void *);
static void wl_get_essid(struct ndis_softc*, void *);
static int wl_set_bssid(struct ndis_softc*, const void *);
static void wl_get_bssid(struct ndis_softc*, void *);
static int wl_set_bsstype(struct ndis_softc*, const void *);
static void wl_get_bsstype(struct ndis_softc*, void *);
static void wl_get_linkstatus(struct ndis_softc*, void *);
static int wl_set_authmode(struct ndis_softc*, const void *);
static void wl_get_authmode(struct ndis_softc*, void *);
static int wl_set_encrypt(struct ndis_softc*, const void *);
static void wl_get_encrypt(struct ndis_softc*, void *);
static void wl_get_rssi(struct ndis_softc*, void *);
static int wl_set_phy(struct ndis_softc*, const void *);
static int wl_get_phy(struct ndis_softc*, void *);
static void wl_get_esslist(struct ndis_softc*, void *);
static int wl_set_wepkey(struct ndis_softc*, const void *);
#if BRUSSELSSUPPORT
static void wl_get_desrates(struct ndis_softc*, void *);
static int wl_set_desrates(struct ndis_softc*, const void *);
static void wl_get_capability(struct ndis_softc*, void *);
static int wl_set_wpa(struct ndis_softc*, const void *);
static void wl_get_wpa(struct ndis_softc*, void *);
static void wl_get_scanresults(struct ndis_softc*, void *);
static int wl_set_optie(struct ndis_softc*, const void *);
static int wl_set_delkey(struct ndis_softc*, const void *);
static int wl_set_mlme(struct ndis_softc*, const void *);
static int wl_set_wpakey(struct ndis_softc*, const void *);
#endif
static void wl_get_suprates(struct ndis_softc*, void *);

#if BRUSSELSSUPPORT
#define	IEEE80211_LOCK(_ic)		\
	mutex_enter(&(_ic)->ic_genlock)
#define	IEEE80211_UNLOCK(_ic)		\
	mutex_exit(&(_ic)->ic_genlock)
#define	IEEE80211_NODE_LOCK(_nt)		\
	mutex_enter(&(_nt)->nt_nodelock)
#define	IEEE80211_NODE_UNLOCK(_nt)		\
	mutex_exit(&(_nt)->nt_nodelock)
#define	IEEE80211_RV(v)			((v) & IEEE80211_RATE_VAL)
#endif

static size_t
wifi_strnlen(const char *s, size_t n)
{
	size_t i;

	for (i = 0; i < n && s[i] != '\0'; i++)
		/* noop */;
	return (i);
}

/*
 * The following routines are for brussels support
 */

/*
 * MAC_PROP_WL_ESSID
 */
static int
wl_set_essid(struct ndis_softc *sc, const void *wldp_buf)
{
	int rval, len, err = 0;
	char *essid;
	wl_essid_t *iw_essid = (wl_essid_t *)wldp_buf;
	struct ndis_80211_ssid win_ssid;
	struct ieee80211com *ic = &sc->ic;
	
	NDISPROP((CE_CONT,"ndis wl_set_essid\n"));
	if (iw_essid->wl_essid_length > IEEE80211_NWID_LEN) {
		NDISPROP((CE_WARN, "wl_set_essid: "
		    "essid too long, %u, max %u\n",
		    iw_essid->wl_essid_length, IEEE80211_NWID_LEN));
		err = EINVAL;
		return (err);
	}

	essid = iw_essid->wl_essid_essid;
	essid[IEEE80211_NWID_LEN] = 0;

	NDISPROP((CE_CONT, "wl_set_essid: "
	    "set essid=%s length=%d\n",
	    essid, iw_essid->wl_essid_length));

	ic->ic_des_esslen = iw_essid->wl_essid_length;
	if (ic->ic_des_esslen != 0)
		bcopy(essid, ic->ic_des_essid, ic->ic_des_esslen);
	if (ic->ic_des_esslen < IEEE80211_NWID_LEN)
		ic->ic_des_essid[ic->ic_des_esslen] = 0;

	err = ENETRESET;
	len = sizeof(ndis_80211_ssid);
	bzero((char *)&win_ssid, len);
	win_ssid.ns_ssidlen = ic->ic_des_esslen;
	if (win_ssid.ns_ssidlen == 0)
		win_ssid.ns_ssidlen = 1;
	else
		bcopy(ic->ic_des_essid, win_ssid.ns_ssid, win_ssid.ns_ssidlen);
	rval = ndis_set_info(sc, OID_802_11_SSID, (void *)&win_ssid, &len);
	if (rval)
		err = EINVAL;
	return (err);
}

static void
wl_get_essid(struct ndis_softc *sc, void *wldp_buf)
{
	char *essid;
	wl_essid_t ow_essid;
	struct ieee80211com *ic = &sc->ic;

	essid = (char *)ic->ic_des_essid;
	if (essid[0] == '\0')
		essid = (char *)ic->ic_bss->in_essid;

	NDISPROP((CE_CONT,"ndis wl_get_essid\n"));
	bzero(&ow_essid, sizeof (wl_essid_t));
	ow_essid.wl_essid_length = wifi_strnlen((const char *)essid,
	    IEEE80211_NWID_LEN);
	bcopy(essid, ow_essid.wl_essid_essid,
	    ow_essid.wl_essid_length);
	bcopy(&ow_essid, wldp_buf, sizeof (wl_essid_t));

}

/*
 * MAC_PROP_WL_BSSID
 */
static int
wl_set_bssid(struct ndis_softc *sc, const void* wldp_buf)
{
	int len, rval;
	char *bssid = (char *)wldp_buf;
	struct ieee80211com *ic = &sc->ic;
	
	NDISPROP((CE_CONT, "wl_set_bssid: "
	    "set bssid=%s\n",
	    ieee80211_macaddr_sprintf(wldp_buf)));
	len = 6;
	rval = ndis_set_info(sc, OID_802_11_BSSID, (void *)bssid, &len);
	bcopy(wldp_buf, ic->ic_des_bssid, sizeof (wl_bssid_t));
	ic->ic_flags |= IEEE80211_F_DESBSSID;

	return (ENETRESET);
}

static void
wl_get_bssid(struct ndis_softc *sc, void *wldp_buf)
{
	uint8_t bssid[6];
	int len;
	struct ieee80211com *ic = &sc->ic;

	NDISPROP((CE_CONT,"ndis wl_get_bssid\n"));
	if (ic->ic_flags & IEEE80211_F_DESBSSID)
		bcopy(ic->ic_des_bssid, bssid, 6);
	else{
		len = 6;
		(void) ndis_get_info(sc, OID_802_11_BSSID, (void *)bssid, &len);
	}
	bcopy(bssid, wldp_buf, sizeof (wl_bssid_t));
}

/*
 * MAC_PROP_WL_BSSTYP
 */
static int
wl_set_bsstype(struct ndis_softc *sc, const void *wldp_buf)
{
	int rval, arg, len, err = 0;
	wl_bss_type_t *iw_opmode = (wl_bss_type_t *)wldp_buf;
	struct ieee80211com *ic = &sc->ic;
	
	NDISPROP((CE_CONT, "wl_set_bsstype: "
	    "set bsstype=%u\n", *iw_opmode));

	switch (*iw_opmode) {
	case WL_BSS_BSS:
		ic->ic_flags &= ~IEEE80211_F_IBSSON;
		ic->ic_opmode = IEEE80211_M_STA;
		err = ENETRESET;
		break;
	case WL_BSS_IBSS:
		if ((ic->ic_caps & IEEE80211_C_IBSS) == 0) {
			err = ENOTSUP;
			break;
		}

		ic->ic_opmode = IEEE80211_M_IBSS;
		err = ENETRESET;
		break;
	default:
		NDISPROP((CE_WARN, "wl_set_bsstype: "
		    "unknown opmode\n"));
		err = EINVAL;
		break;
	}

	switch (*iw_opmode) {
	case WL_BSS_IBSS:
		arg = NDIS_80211_NET_INFRA_IBSS;
		break;
	case WL_BSS_BSS: 
		arg = NDIS_80211_NET_INFRA_BSS;
		break;
	default:
		return (EINVAL);
	}
	
	len = sizeof(arg);
	rval = ndis_set_info(sc, OID_802_11_INFRASTRUCTURE_MODE, 
				(void *)&arg, &len);
	if(rval)
		err = EINVAL;
	
	return (err);
}

static void
wl_get_bsstype(struct ndis_softc *sc, void *wldp_buf)
{
	wl_bss_type_t ow_opmode;
	int len, rval, arg;
	struct ieee80211com *ic = &sc->ic;
	
	NDISPROP((CE_CONT,"ndis wl_get_bsstype\n"));
	len = sizeof(arg);
	rval = ndis_get_info(sc, OID_802_11_INFRASTRUCTURE_MODE, 
			&arg, &len);

	switch (arg) {
	case NDIS_80211_NET_INFRA_IBSS:
		ic->ic_opmode = IEEE80211_M_IBSS;
		break;
	case NDIS_80211_NET_INFRA_BSS:
		ic->ic_opmode = IEEE80211_M_STA;
		break;
	case NDIS_80211_NET_INFRA_AUTO:
	default:
		ic->ic_opmode = 0x4;
		break;
	}
	
	switch (ic->ic_opmode) {
	case IEEE80211_M_STA:
		ow_opmode = WL_BSS_BSS;
		break;
	case IEEE80211_M_IBSS:
		ow_opmode = WL_BSS_IBSS;
		break;
	default:
		ow_opmode = WL_BSS_ANY;
		break;
	}

	bcopy(&ow_opmode, wldp_buf, sizeof (wl_bss_type_t));
}

/*
 * MAC_PROP_WL_LINKSTATUS
 */
static void
wl_get_linkstatus(struct ndis_softc *sc, void *wldp_buf)
{
	wl_linkstatus_t ow_linkstat;
	struct ieee80211com *ic = &sc->ic;

	NDISPROP((CE_CONT,"ndis wl_get_linkstatus\n"));
	ow_linkstat = (ic->ic_state == IEEE80211_S_RUN) ?
	    WL_CONNECTED : WL_NOTCONNECTED;

	bcopy(&ow_linkstat, wldp_buf, sizeof (wl_linkstatus_t));
}

/*
 * MAC_PROP_AUTH_MODE
 */
static int
wl_set_authmode(struct ndis_softc *sc, const void *wldp_buf)
{
	int rval, arg, len, err = 0;
	wl_authmode_t *iw_auth = (wl_authmode_t *)wldp_buf;
	struct ieee80211com *ic = &sc->ic;
	
	NDISPROP((CE_CONT,"ndis wl_set_authmode"));

	switch (*iw_auth) {
	case WL_OPENSYSTEM:
		arg = NDIS_80211_AUTHMODE_OPEN;	
		ic->ic_bss->in_authmode = *iw_auth;
		err = ENETRESET;		
		break;
	case WL_SHAREDKEY:
		arg = NDIS_80211_AUTHMODE_SHARED;			
		ic->ic_bss->in_authmode = *iw_auth;
		err = ENETRESET;
		break;
	default:
		cmn_err(CE_WARN, "wl_set_authmode: "
		    "unknown authmode %u\n", *iw_auth);
		err = EINVAL;
		return (err);
	}

	len = sizeof(arg);
	rval = ndis_set_info(sc, OID_802_11_AUTHENTICATION_MODE, 
			(void *)&arg, &len);
	if (rval){
		cmn_err(CE_WARN,"ndis set authmode failed\n");
		err = EINVAL;
	}
	return (err);
}

static void
wl_get_authmode(struct ndis_softc *sc, void *wldp_buf)
{
	wl_authmode_t ow_auth;
	int len, rval, arg;
	struct ieee80211com *ic = &sc->ic;
	
	NDISPROP((CE_CONT,"ndis wl_get_authmode\n"));
	len = sizeof(arg);
	rval = ndis_get_info(sc, OID_802_11_AUTHENTICATION_MODE,
			&arg, &len);
	switch (arg) {
	case NDIS_80211_AUTHMODE_OPEN: 
		ic->ic_bss->in_authmode = WL_OPENSYSTEM;
		break;
	case NDIS_80211_AUTHMODE_SHARED:
		ic->ic_bss->in_authmode = WL_SHAREDKEY;
		break;
	case NDIS_80211_AUTHMODE_WPA:
	case NDIS_80211_AUTHMODE_WPANONE:
	case NDIS_80211_AUTHMODE_WPAPSK:
		cmn_err(CE_WARN,"wpa TBD\n");
		break;
	default:
		cmn_err(CE_WARN,"unknown auth mode\n");
		break;
	}

	ow_auth = ic->ic_bss->in_authmode;
	bcopy(&ow_auth, wldp_buf, sizeof (wl_authmode_t));
}

/*
 * MAC_PROP_WL_ENCRYPTION
 */
static int
wl_set_encrypt(struct ndis_softc *sc, const void *wldp_buf)
{
	int err = 0;
	uint32_t flags;
	wl_encryption_t *iw_encryp = (wl_encryption_t *)wldp_buf;
	struct ieee80211com *ic = &sc->ic;
	int arg, len, rval;
	
	flags = ic->ic_flags;
	if (*iw_encryp == WL_NOENCRYPTION)
		flags &= ~IEEE80211_F_PRIVACY;
	else
		flags |= IEEE80211_F_PRIVACY;

	NDISPROP((CE_CONT,"ndis wl_set_encrypt, flag %d\n", flags));
	if (ic->ic_flags != flags) {
		ic->ic_flags = flags;
		err = ENETRESET;
	}

	if (ic->ic_flags & IEEE80211_F_PRIVACY) {
		arg = NDIS_80211_WEPSTAT_ENABLED;
		len = sizeof(arg);
		rval = ndis_set_info(sc, OID_802_11_WEP_STATUS, (void *)&arg, &len);
	} else {
		arg = NDIS_80211_WEPSTAT_DISABLED;
		len = sizeof(arg);
		rval = ndis_set_info(sc, OID_802_11_WEP_STATUS, (void *)&arg, &len);
	}

	if(rval){
		cmn_err(CE_CONT,"ndis wl_set_encrypt flags %d failed\n", flags);
		err = EINVAL;
	}
	return (err);
}

static void
wl_get_encrypt(struct ndis_softc *sc, void *wldp_buf)
{
	wl_encryption_t *ow_encryp;
	struct ieee80211com *ic = &sc->ic;

	NDISPROP((CE_CONT,"ndis wl_get_encrypt\n"));
	ow_encryp = (wl_encryption_t *)wldp_buf;
	*ow_encryp = (ic->ic_flags & IEEE80211_F_PRIVACY) ? 1 : 0;
	if (ic->ic_flags & IEEE80211_F_WPA)
		*ow_encryp = WL_ENC_WPA;
}

/*
 * MAC_PROP_WL_RSSI
 */
static void
wl_get_rssi(struct ndis_softc *sc, void *wldp_buf)
{
	wl_rssi_t *ow_rssi;
	struct ieee80211com *ic = &sc->ic;

	NDISPROP((CE_CONT,"ndis wl_get_rssi\n"));
	ow_rssi = (wl_rssi_t *)wldp_buf;
	*ow_rssi = sc->ndis_rssi;
}

/*
 * MAC_PROP_WL_PHY_CONFIG
 */

static int
wl_set_phy(struct ndis_softc *sc, const void* wldp_buf)
{
	int len, rval, err = 0;
	int16_t ch;
	wl_dsss_t *dsss;
	wl_phy_conf_t *iw_phy = (wl_phy_conf_t *)wldp_buf;
	struct ndis_80211_config config;
	struct ieee80211com *ic = &sc->ic;
	int chanflag = 0;

	dsss = (wl_dsss_t *)iw_phy;
	ch = dsss->wl_dsss_channel;

	NDISPROP((CE_CONT,"ndis wl_set_phy to chan %d\n", ch));
	if (ch < 1 || ch > IEEE80211_CHAN_MAX) 
		return EINVAL;

	len = sizeof(config);
	bzero((char *)&config, len);
	config.nc_length = len;
	config.nc_fhconfig.ncf_length = sizeof(ndis_80211_config_fh);
	rval = ndis_get_info(sc, OID_802_11_CONFIGURATION, &config, &len); 
	if (rval)
		return EINVAL;

	/*
	 * Some drivers expect us to initialize these values, so
	 * provide some defaults.
	 */
	if (config.nc_beaconperiod == 0)
		config.nc_beaconperiod = 100;
	if (config.nc_atimwin == 0)
		config.nc_atimwin = 100;
	if (config.nc_fhconfig.ncf_dwelltime == 0)
		config.nc_fhconfig.ncf_dwelltime = 200;

	if (ic->ic_sup_rates[IEEE80211_MODE_11G].ir_nrates) /* IEEE80211_CHAN_G */
		chanflag |= IEEE80211_CHAN_2GHZ|IEEE80211_CHAN_DYN;
	if (ch <= 14)	/* IEEE80211_CHAN_B */
		chanflag |= IEEE80211_CHAN_2GHZ|IEEE80211_CHAN_CCK;
	if (ic->ic_sup_rates[IEEE80211_MODE_11A].ir_nrates && ch > 14)	/* IEEE80211_CHAN_A */
		chanflag |= IEEE80211_CHAN_5GHZ|IEEE80211_CHAN_OFDM;
		
	ic->ic_ibss_chan = &ic->ic_sup_channels[ch]; //fix channel pointer
	ic->ic_ibss_chan->ich_freq = ieee80211_ieee2mhz(ch, chanflag);
	config.nc_dsconfig = ic->ic_ibss_chan->ich_freq * 1000;
	ic->ic_bss->in_chan = ic->ic_ibss_chan;

	len = sizeof(config);
	config.nc_length = len;
	config.nc_fhconfig.ncf_length = sizeof(ndis_80211_config_fh);
	rval = ndis_set_info(sc, OID_802_11_CONFIGURATION, (void *)&config, &len);
	if (rval){
		cmn_err(CE_CONT,"ndis set phy[%d] failed\n", ch);
		err = EINVAL;
	}
	return (err);
}

static int
wl_get_phy(struct ndis_softc *sc, void *wldp_buf)
{
	int rval, len, err = 0;
	wl_phy_conf_t *ow_phy;
	struct ieee80211com *ic = &sc->ic;
	struct ieee80211_channel *ch = ic->ic_curchan;
	ndis_80211_config config;
	
	NDISPROP((CE_CONT,"ndis wl_get_phy\n"));
	ow_phy = (wl_phy_conf_t *)wldp_buf;
	bzero(wldp_buf, sizeof (wl_phy_conf_t));

	len = sizeof(config);
	bzero((char *)&config, len);
	config.nc_length = len;
	config.nc_fhconfig.ncf_length = sizeof(ndis_80211_config_fh);
	rval = ndis_get_info(sc, OID_802_11_CONFIGURATION, &config, &len); 
	if (rval)
		return (EINVAL);

	if (config.nc_fhconfig.ncf_length){
		wl_fhss_t *fhss = (wl_fhss_t *)ow_phy;
		fhss->wl_fhss_subtype = WL_FHSS;
		fhss->wl_fhss_channel = ieee80211_mhz2ieee(config.nc_dsconfig / 1000, 0);
		return (err);
	}

	/* XXX this needs to be improved by OID_802_11_NETWORK_TYPE_IN_USE */
	if ((config.nc_dsconfig >= 2412000 && config.nc_dsconfig <= 2484000) ||
		(config.nc_dsconfig >= 5000000 && config.nc_dsconfig <= 6000000)){
		wl_erp_t *erp = (wl_erp_t *)ow_phy;
		erp->wl_erp_subtype = WL_ERP;
		erp->wl_erp_channel = ieee80211_mhz2ieee(config.nc_dsconfig / 1000, 0);
		return (err);
	}
	
	return (EINVAL);
}

extern void ndis_wpa_ie_parser(wl_ess_conf_t *ap, uint8_t *ie,
 				    size_t ie_len);
extern int ndis_nettype_parser(wl_ess_conf_t *ap, ndis_wlan_bssid_ex *wb);

int ap_list_len;
char ap_list[8192];
/*
 * MAC_PROP_WL_ESS_LIST
 */
static void
wl_get_esslist(struct ndis_softc *sc, void *wldp_buf)
{
	wl_ess_list_t *req_ess_list;
	struct ieee80211com *ic = &sc->ic;
	uint32_t i,j, rval;
	int len; /* was uint32_t */
	uint8_t *db;
	ndis_80211_bssid_list_ex *ess_list;
	ndis_wlan_bssid_ex *wb;
	int rssi; /* was uint32_t */

	req_ess_list = (wl_ess_list_t *)wldp_buf;
	if(sc->ndis_link){
		bcopy(ap_list, req_ess_list, ap_list_len);
		return;
	}
	req_ess_list->wl_ess_list_num = 0;
	
	len = 8192;
	ess_list = (ndis_80211_bssid_list_ex *)
			kmem_zalloc(len, KM_NOSLEEP);
	if (ess_list == NULL)
		return;
	rval = ndis_get_info(sc, OID_802_11_BSSID_LIST, ess_list, &len);
	NDISPROP((CE_CONT,"list len=%d, rval=%x\n",len,rval));

	wb = (ndis_wlan_bssid_ex *)&ess_list->nblx_bssid[0];

	for(i=0; i< ess_list->nblx_items; i++){
		NDISPROP((CE_CONT,"scan item %d\n",i));
		wl_ess_conf_t *p_ess_conf;
		p_ess_conf = &req_ess_list->wl_ess_list_ess[i];
		p_ess_conf->wl_ess_conf_length = sizeof (struct wl_ess_conf);
		db = wb->nwbx_ssid.ns_ssid;
		NDISPROP((CE_CONT,"SSID name %s\n",wb->nwbx_ssid.ns_ssid));
		bcopy(wb->nwbx_ssid.ns_ssid, p_ess_conf->wl_ess_conf_essid.wl_essid_essid, 32);
		p_ess_conf->wl_ess_conf_essid.wl_essid_length = wb->nwbx_ssid.ns_ssidlen;
		bcopy(wb->nwbx_macaddr, p_ess_conf->wl_ess_conf_bssid, 6);
		if(wb->nwbx_netinfra == NDIS_80211_NET_INFRA_IBSS)
			p_ess_conf->wl_ess_conf_bsstype = WL_BSS_IBSS;
		else if(wb->nwbx_netinfra == NDIS_80211_NET_INFRA_BSS)
			p_ess_conf->wl_ess_conf_bsstype = WL_BSS_BSS;
		p_ess_conf->wl_ess_conf_wepenabled = 
			(wb->nwbx_privacy)?WL_ENC_WEP : WL_NOENCRYPTION;
		NDISPROP((CE_CONT, "wb->nwbx_privacy(none/wep) = %d\n", wb->nwbx_privacy));
		//wpa detection
		ndis_wpa_ie_parser(p_ess_conf, (uint8_t *)&wb->nwbx_ies[0], wb->nwbx_ielen);
		/*signal strength*/
		rssi = (wb->nwbx_rssi + 200)/12;
		if(rssi < 0)
			rssi = 0;
		else if(rssi > 15)
			rssi = 15;			
		p_ess_conf->wl_ess_conf_sl = rssi;
		//AP maxrate detect
		for (j = 0; j < sizeof(wb->nwbx_supportedrates); j++) {
			if ((wb->nwbx_supportedrates[j] & 0x7f) >
		       			p_ess_conf->wl_supported_rates[0]) {
				p_ess_conf->wl_supported_rates[0] =
					wb->nwbx_supportedrates[j] & 0x7f;
	                 }
	        }
		//nettype
		NDISPROP((CE_CONT,"wb->nwbx_nettype = 0x%x\n",wb->nwbx_nettype));
		ndis_nettype_parser(p_ess_conf, wb);
		wb = (ndis_wlan_bssid_ex *)((char *)wb + wb->nwbx_len);
	}
	
	req_ess_list->wl_ess_list_num = ess_list->nblx_items;
	ap_list_len = 4 + (sizeof(wl_ess_conf_t) * ess_list->nblx_items);
	bcopy(req_ess_list, ap_list, ap_list_len); 
	kmem_free(ess_list, 8192);
}

/*
 * MAC_PROP_WL_WEP_KEY
 */
static int
wl_set_wepkey(struct ndis_softc *sc, const void *wldp_buf)
{
	int	 rval=0, err = 0;
	uint16_t i;
	int len; /* was uint32_t */
	uint32_t klen;
	wl_wep_key_t *wepkey = (wl_wep_key_t *)wldp_buf;
	ndis_80211_wep	wep;
	struct ieee80211com *ic = &sc->ic;
	
	NDISPROP((CE_CONT,"ndis wl_set_wepkey\n"));
	/* set all valid keys */
	for (i = 0; i < MAX_NWEPKEYS; i++) {
		if (wepkey[i].wl_wep_operation != WL_ADD)
			continue;
		klen = wepkey[i].wl_wep_length;
		/* backup the key info*/
		ic->ic_nw_keys[i].wk_keylen = klen;
		bcopy(wepkey[i].wl_wep_key, ic->ic_nw_keys[i].wk_key, klen);
		
		if (klen > IEEE80211_KEYBUF_SIZE) {
			NDISPROP((CE_WARN, "wl_set_wepkey: "
			    "invalid wepkey length, %u\n", klen));
			err = EINVAL;
			continue;  /* continue to set other keys */
		}
		if (klen == 0)
			continue;

		/*
		 *  Set key contents. Only WEP is supported
		 */
		bzero((char *)&wep, sizeof(wep));
		wep.nw_keylen = klen;
		wep.nw_keyidx = i;
		wep.nw_length = (sizeof(uint32_t) * 3) + wep.nw_keylen;
		if (i == (ic->ic_crypto.cs_def_txkey))
			wep.nw_keyidx |= NDIS_80211_WEPKEY_TX;
		else
			wep.nw_keyidx &= ~NDIS_80211_WEPKEY_TX;
		bcopy(wepkey[i].wl_wep_key, wep.nw_keydata, wep.nw_keylen);
		len = sizeof(wep);
		rval = ndis_set_info(sc, OID_802_11_ADD_WEP, &wep, &len);	
		if (rval)
			err = EINVAL;		
	}
	if (err == 0)
		err = ENETRESET;

	return (err);
}

#if BRUSSELSSUPPORT
/*
 * MAC_PROP_WL_SUP_RATE
 */
static void
wl_get_suprates(struct ndis_softc *sc, void *wldp_buf)
{
	int i, j;
	unsigned int k, l;
	uint8_t srates;
	uint8_t *drates;
	wl_rates_t *wl_rates;
	const struct ieee80211_rateset *srs;
	struct ieee80211com *ic = &sc->ic;

	wl_rates = (wl_rates_t *)wldp_buf;

	wl_rates->wl_rates_num = 0;
	drates = (uint8_t *)wl_rates->wl_rates_rates;
	for (i = 0; i < IEEE80211_MODE_MAX; i++) {
		srs = &ic->ic_sup_rates[i];
		if (srs->ir_nrates == 0)
			continue;
		for (j = 0; j < srs->ir_nrates; j++) {
			srates = IEEE80211_RV(srs->ir_rates[j]);
			/* sort & skip duplicated rates */
			for (k = 0; k < wl_rates->wl_rates_num; k++) {
				if (srates <= drates[k])
					break;
			}
			if (srates == drates[k])
				/* skip duplicated rates */
				continue;
			/* sort */
			for (l = wl_rates->wl_rates_num; l > k; l--)
				drates[l] = drates[l-1];
			drates[k] = srates;
			wl_rates->wl_rates_num++;
		}
	}

}


/*
 * MAC_PROP_WL_DESIRED_RATESa
 */
static int
wl_set_desrates(struct ndis_softc *sc, const void *wldp_buf)
{
	int err = 0;
	int i, j;
	uint8_t drate;
	boolean_t isfound;
	wl_rates_t *iw_rates = (wl_rates_t *)wldp_buf;
	struct ieee80211com *ic = &sc->ic;
	struct ieee80211_node *in = ic->ic_bss;
	struct ieee80211_rateset *rs = &in->in_rates;

	drate = iw_rates->wl_rates_rates[0];
	if (ic->ic_fixed_rate == drate)
		return (err);

	if (drate == 0) {
		ic->ic_fixed_rate = IEEE80211_FIXED_RATE_NONE;
		if (ic->ic_state == IEEE80211_S_RUN) {
			IEEE80211_UNLOCK(ic);
			ieee80211_new_state(ic, IEEE80211_S_ASSOC, 0);
			IEEE80211_LOCK(ic);
		}
		return (err);
	}

	/*
	 * Set desired rate. The desired rate is for data transfer
	 * and usally is checked and used when driver changes to
	 * RUN state.
	 * If the driver is in AUTH | ASSOC | RUN state, desired
	 * rate is checked anainst rates supported by current ESS.
	 * If it's supported and current state is AUTH|ASSOC, nothing
	 * needs to be done by driver since the desired rate will
	 * be enabled when the device changes to RUN state. And
	 * when current state is RUN, Re-associate with the ESS to
	 * enable the desired rate.
	 */

	if (ic->ic_state != IEEE80211_S_INIT &&
	    ic->ic_state != IEEE80211_S_SCAN) {
		for (i = 0; i < rs->ir_nrates; i++) {
			if (drate == IEEE80211_RV(rs->ir_rates[i]))
				break;
		}
		/* supported */
		if (i < rs->ir_nrates) {
			ic->ic_fixed_rate = drate;
			if (ic->ic_state == IEEE80211_S_RUN) {
				IEEE80211_UNLOCK(ic);
				ieee80211_new_state(ic,
				    IEEE80211_S_ASSOC, 0);
				IEEE80211_LOCK(ic);
			}
			return (err);
		}
	}

	/*
	 * In INIT or SCAN state
	 * check if the desired rate is supported by device
	 */
	isfound = B_FALSE;
	for (i = 0; i < IEEE80211_MODE_MAX; i++) {
		rs = &ic->ic_sup_rates[i];
		for (j = 0; j < rs->ir_nrates; j++) {
			if (drate ==  IEEE80211_RV(rs->ir_rates[j])) {
				isfound = B_TRUE;
				break;
			}
		}
		if (isfound)
			break;
	}
	if (!isfound) {
		cmn_err(CE_WARN, "wl_set_desrates: "
		    "invald rate %d\n", drate);
		err = EINVAL;
		return (err);
	}
	ic->ic_fixed_rate = drate;
	if (ic->ic_state != IEEE80211_S_SCAN)
		err = ENETRESET;

	return (err);
}

static void
wl_get_desrates(struct ndis_softc *sc, void *wldp_buf)
{
	uint8_t srate;
	wl_rates_t ow_rates;
	struct ieee80211com *ic = &sc->ic;
	struct ieee80211_node *in = ic->ic_bss;
	struct ieee80211_rateset *rs = &in->in_rates;

	srate = rs->ir_rates[in->in_txrate] & IEEE80211_RATE_VAL;
	ow_rates.wl_rates_num = 1;
	ow_rates.wl_rates_rates[0] =
	    (ic->ic_fixed_rate == IEEE80211_FIXED_RATE_NONE) ?
	    srate : ic->ic_fixed_rate;
	bcopy(&ow_rates, wldp_buf, sizeof (wl_rates_t));

}


/*
 * MAC_PROP_WL_CAPABILITY
 */
static void
wl_get_capability(struct ndis_softc *sc, void *wldp_buf)
{
	wl_capability_t ow_caps;
	struct ieee80211com *ic = &sc->ic;

	ow_caps.caps = ic->ic_caps;
	bcopy(&ow_caps, wldp_buf, sizeof (wl_capability_t));

}

/*
 * MAC_PROP_WL_WPA
 */
static int
wl_set_wpa(struct ndis_softc *sc, const void *wldp_buf)
{
	int err = 0;
	wl_wpa_t *wpa = (wl_wpa_t *)wldp_buf;
	struct ieee80211com *ic = &sc->ic;

	if (wpa->wpa_flag > 0) {
		/* enable wpa mode */
		ic->ic_flags |= IEEE80211_F_PRIVACY;
		ic->ic_flags |= IEEE80211_F_WPA;
	} else {
		ic->ic_flags &= ~IEEE80211_F_PRIVACY;
		ic->ic_flags &= ~IEEE80211_F_WPA;
	}

	return (err);
}

static void
wl_get_wpa(struct ndis_softc *sc, void *wldp_buf)
{
	wl_wpa_t *wpa;
	struct ieee80211com *ic = &sc->ic;

	wpa = (wl_wpa_t *)wldp_buf;
	wpa->wpa_flag = ((ic->ic_flags & IEEE80211_F_WPA) ? 1 : 0);

}

/*
 * MAC_PROP_WL_SCANRESULTS
 */

static void
wl_get_scanresults(struct ndis_softc *sc, void *wldp_buf)
{
	wl_wpa_ess_t *sr;
	struct ieee80211com *ic = &sc->ic;
	ieee80211_node_t *in;
	ieee80211_node_table_t *nt;
	int ap_num;
	int len;

	sr = (wl_wpa_ess_t *)wldp_buf;
	sr->count = 0;
	ap_num = 0;

	nt = &ic->ic_scan;
	IEEE80211_NODE_LOCK(nt);
	in = list_head(&nt->nt_node);

	while (in != NULL) {
		/* filter out non-wpa APs */
		if (in->in_wpa_ie == NULL) {
			in = list_next(&nt->nt_node, in);
			continue;
		}
		bcopy(in->in_bssid, sr->ess[ap_num].bssid,
		    IEEE80211_ADDR_LEN);
		sr->ess[ap_num].ssid_len = in->in_esslen;
		bcopy(in->in_essid, sr->ess[ap_num].ssid,
		    in->in_esslen);
		sr->ess[ap_num].freq = in->in_chan->ich_freq;

		len = in->in_wpa_ie[1] + 2;
		bcopy(in->in_wpa_ie, sr->ess[ap_num].wpa_ie, len);
		sr->ess[ap_num].wpa_ie_len = len;

		ap_num++;
		in = list_next(&nt->nt_node, in);
	}
	IEEE80211_NODE_UNLOCK(nt);
	sr->count = ap_num;

}

/*
 * MAC_PROP_WL_SETOPTIE
 */
static int
wl_set_optie(struct ndis_softc *sc, const void *wldp_buf)
{
	int err = 0;
	char *ie;
	struct ieee80211com *ic = &sc->ic;
	wl_wpa_ie_t *ie_in = (wl_wpa_ie_t *)wldp_buf;

	if (ic->ic_opmode != IEEE80211_M_STA) {
		cmn_err(CE_WARN, "wl_set_optie: opmode err\n");
		err = EINVAL;
		return (err);
	}
	if (ie_in->wpa_ie_len > IEEE80211_MAX_OPT_IE) {

		cmn_err(CE_WARN, "wl_set_optie: optie is too long\n");

		err = EINVAL;
		return (err);
	}

	ie = ieee80211_malloc(ie_in->wpa_ie_len);
	(void) memcpy(ie, ie_in->wpa_ie, ie_in->wpa_ie_len);
	if (ic->ic_opt_ie != NULL) {
		ieee80211_free(ic->ic_opt_ie);
	}
	ic->ic_opt_ie = ie;
	ic->ic_opt_ie_len = ie_in->wpa_ie_len;

	return (err);
}

/*
 * MAC_PROP_WL_DELKEY
 */
static int
wl_set_delkey(struct ndis_softc *sc, const void *wldp_buf)
{
	int err = 0;
	unsigned int kid;
	struct ieee80211com *ic = &sc->ic;
	wl_del_key_t *dk = (wl_del_key_t *)wldp_buf;

	kid = dk->idk_keyix;

	if (kid == IEEE80211_KEYIX_NONE ||
	    kid >= IEEE80211_WEP_NKID) {
		cmn_err(CE_WARN, "wl_set_delkey: incorrect keyix\n");
		err = EINVAL;
		return (err);
	} else {
		(void) ieee80211_crypto_delkey(ic,
		    &ic->ic_nw_keys[kid]);
		ieee80211_mac_update(ic);
	}

	return (err);
}

/*
 * MAC_PROP_WL_MLME
 */

static int
wl_set_mlme(struct ndis_softc *sc, const void *wldp_buf)
{
	int err = 0;
	uint32_t flags;
	ieee80211_node_t *in;
	struct ieee80211com *ic = &sc->ic;
	wl_mlme_t *mlme = (wl_mlme_t *)wldp_buf;

	switch (mlme->im_op) {
	case IEEE80211_MLME_DISASSOC:
	case IEEE80211_MLME_DEAUTH:
		if (ic->ic_opmode == IEEE80211_M_STA) {
			/*
			 * Mask ic_flags of IEEE80211_F_WPA to disable
			 * ieee80211_notify temporarily.
			 */
			flags = ic->ic_flags;
			ic->ic_flags &= ~IEEE80211_F_WPA;

			IEEE80211_UNLOCK(ic);
			ieee80211_new_state(ic, IEEE80211_S_INIT,
			    mlme->im_reason);
			IEEE80211_LOCK(ic);

			ic->ic_flags = flags;
		}
		break;
	case IEEE80211_MLME_ASSOC:
		if (ic->ic_opmode != IEEE80211_M_STA) {
			cmn_err(CE_WARN, "wifi_cfg_setmlme: opmode err\n");
			err = EINVAL;
			break;
		}
		if (ic->ic_des_esslen != 0) {
		/*
		 * Desired ssid specified; must match both bssid and
		 * ssid to distinguish ap advertising multiple ssid's.
		 */
			in = ieee80211_find_node_with_ssid(&ic->ic_scan,
			    mlme->im_macaddr,
			    ic->ic_des_esslen,
			    ic->ic_des_essid);
		} else {
		/*
		 * Normal case; just match bssid.
		 */
			in = ieee80211_find_node(&ic->ic_scan,
			    mlme->im_macaddr);
		}
		if (in == NULL) {
			cmn_err(CE_WARN, "wifi_cfg_setmlme: "
			    "no matched node\n");
			err = EINVAL;
			break;
		}
		IEEE80211_UNLOCK(ic);
		ieee80211_sta_join(ic, in);
		IEEE80211_LOCK(ic);
		break;
	default:
		err = EINVAL;
		break;
	}

	return (err);
}

/*
 * MAC_PROP_WL_WPA_KEY
 */
static int
wl_set_wpakey(struct ndis_softc *sc, const void *wldp_buf)
{
	int err = 0;
	uint16_t kid;
	struct ieee80211com *ic = &sc->ic;
	struct ieee80211_node *in;
	struct ieee80211_key *wk;
	wl_key_t ik;

	bcopy(wldp_buf, &ik, sizeof (wl_key_t));

	/*
	 * cipher support is verified by ieee80211_crypt_newkey
	 * this also checks ik.ik_keylen > sizeof(wk->wk_key)
	 */
	if (ik.ik_keylen > sizeof (ik.ik_keydata)) {
		cmn_err(CE_WARN, "wl_set_wpakey: key is too long\n");
		err = EINVAL;
		return (err);
	}
	kid = ik.ik_keyix;
	if (kid == IEEE80211_KEYIX_NONE || kid >= IEEE80211_WEP_NKID) {
		cmn_err(CE_WARN, "wl_set_wpakey: incorrect keyix\n");
		err = EINVAL;
		return (err);
	} else {
		wk = &ic->ic_nw_keys[kid];
		/*
		 * Globle slots start off w/o any assigned key index.
		 * Force one here for consistency with WEPKEY.
		 */
		if (wk->wk_keyix == IEEE80211_KEYIX_NONE)
			wk->wk_keyix = kid;
		in = NULL;
	}

	KEY_UPDATE_BEGIN(ic);
	if (ieee80211_crypto_newkey(ic, ik.ik_type,
	    ik.ik_flags, wk)) {
		wk->wk_keylen = ik.ik_keylen;
		/* MIC presence is implied by cipher type */
		if (wk->wk_keylen > IEEE80211_KEYBUF_SIZE)
			wk->wk_keylen = IEEE80211_KEYBUF_SIZE;
		wk->wk_keyrsc = ik.ik_keyrsc;
		wk->wk_keytsc = 0;
		wk->wk_flags |= ik.ik_flags &
		    (IEEE80211_KEY_XMIT | IEEE80211_KEY_RECV);
		(void) memset(wk->wk_key, 0, sizeof (wk->wk_key));
		(void) memcpy(wk->wk_key, ik.ik_keydata,
		    ik.ik_keylen);
		if (!ieee80211_crypto_setkey(ic, wk,
		    in != NULL ? in->in_macaddr : ik.ik_macaddr)) {
			err = EIO;
		} else if ((ik.ik_flags & IEEE80211_KEY_DEFAULT)) {
			ic->ic_def_txkey = kid;
			ieee80211_mac_update(ic);
		}
	} else {
		err = EIO;
	}
	KEY_UPDATE_END(ic);

	return (err);
}
#endif //brussels only

/*
 * Typically invoked by drivers in response to request for
 * information or to change settings from the userland.
 *
 * Return value should be checked by WiFI drivers. Return 0
 * on success. Otherwise, return non-zero value to indicate
 * the error. Driver should operate as below when the return
 * error is:
 * ENETRESET	Reset wireless network and re-start to join a
 * 		WLAN, ENETRESET is returned when a configuration
 * 		parameter has been changed.
 * 		When acknowledge a M_IOCTL message, this error
 * 		is ignored
 */

/* ARGSUSED */
int
ndis_setprop(void *sc_arg, const char *pr_name, mac_prop_id_t wldp_pr_num,
    uint_t wldp_length, const void *wldp_buf)
{
	int err = 0;
	struct ndis_softc *sc = (struct ndis_softc *)sc_arg;
	struct ieee80211com *ic = &sc->ic;

	ASSERT(ic != NULL);
	NDIS_LOCK(sc);

// TODO migrate elsewhere
	if (!ic->ic_tim_len)
		ieee80211_media_init(ic); /* might be needed */
	switch (wldp_pr_num) {
	/* mac_prop_id */
	case MAC_PROP_WL_ESSID:
		err = wl_set_essid(sc, wldp_buf);
		break;
	case MAC_PROP_WL_BSSID:
		err = wl_set_bssid(sc, wldp_buf);
		break;
	case MAC_PROP_WL_PHY_CONFIG:
		err = wl_set_phy(sc, wldp_buf);
		break;
	case MAC_PROP_WL_KEY_TAB:
		err = wl_set_wepkey(sc, wldp_buf);
		break;
	case MAC_PROP_WL_AUTH_MODE:
		err = wl_set_authmode(sc, wldp_buf);
		break;
	case MAC_PROP_WL_ENCRYPTION:
		err = wl_set_encrypt(sc, wldp_buf);
		break;
	case MAC_PROP_WL_BSSTYPE:
		err = wl_set_bsstype(sc, wldp_buf);
		break;
	case MAC_PROP_WL_DESIRED_RATES:
		err = wl_set_desrates(sc, wldp_buf);
		break;
	case MAC_PROP_WL_WPA:
		err = wl_set_wpa(sc, wldp_buf);
		break;
	case MAC_PROP_WL_KEY:
		err = wl_set_wpakey(sc, wldp_buf);
		break;
	case MAC_PROP_WL_DELKEY:
		err = wl_set_delkey(sc, wldp_buf);
		break;
	case MAC_PROP_WL_SETOPTIE:
		err = wl_set_optie(sc, wldp_buf);
		break;
	case MAC_PROP_WL_MLME:
		err = wl_set_mlme(sc, wldp_buf);
		err = ENOTSUP;
		break;
	case MAC_PROP_WL_LINKSTATUS:
	case MAC_PROP_WL_ESS_LIST:
	case MAC_PROP_WL_SUPPORTED_RATES:
	case MAC_PROP_WL_RSSI:
	case MAC_PROP_WL_CAPABILITY:
	case MAC_PROP_WL_SCANRESULTS:
		cmn_err(CE_WARN, "ieee80211_setprop: opmode %d err\n",
						(int)wldp_pr_num);
		err = EINVAL;
		break;
	default:
		cmn_err(CE_WARN, "ieee80211_setprop: opmode %d not supported\n",
						(int)wldp_pr_num);
		err = ENOTSUP;
		break;
	}
	if(err == ENETRESET) err = 0;
	NDIS_UNLOCK(sc);

	return (err);
}

/* ARGSUSED */
int
ndis_getprop(void *sc_arg, const char *pr_name, mac_prop_id_t wldp_pr_num,
    uint_t wldp_length, void *wldp_buf)
{
	int err = 0;
	/* the argument is what was passed as m_driver to mac_register() */
	struct ndis_softc *sc = (struct ndis_softc *)sc_arg;
	struct ieee80211com *ic = &sc->ic;

	if (wldp_length == 0) {
		err = EINVAL;
		return (err);
	}
	bzero(wldp_buf, wldp_length);

	ASSERT(ic != NULL);
	NDIS_LOCK(sc);

// TODO migrate elsewhere
	if (!ic->ic_tim_len)
		ieee80211_media_init(ic); /* might be needed */
	switch (wldp_pr_num) {
	/* mac_prop_id */
	case MAC_PROP_WL_ESSID:
		wl_get_essid(sc, wldp_buf);
		break;
	case MAC_PROP_WL_BSSID:
		wl_get_bssid(sc, wldp_buf);
		break;
	case MAC_PROP_WL_PHY_CONFIG:
		err = wl_get_phy(sc, wldp_buf);
		break;
	case MAC_PROP_WL_AUTH_MODE:
		wl_get_authmode(sc, wldp_buf);
		break;
	case MAC_PROP_WL_ENCRYPTION:
		wl_get_encrypt(sc, wldp_buf);
		break;
	case MAC_PROP_WL_BSSTYPE:
		wl_get_bsstype(sc, wldp_buf);
		break;
	case MAC_PROP_WL_DESIRED_RATES:
		wl_get_desrates(sc, wldp_buf);
		break;
	case MAC_PROP_WL_LINKSTATUS:
		wl_get_linkstatus(sc, wldp_buf);
		break;
	case MAC_PROP_WL_ESS_LIST:
		wl_get_esslist(sc, wldp_buf);
		break;
	case MAC_PROP_WL_SUPPORTED_RATES:
		wl_get_suprates(sc, wldp_buf);
		break;
	case MAC_PROP_WL_RSSI:
		wl_get_rssi(sc, wldp_buf);
		break;
	case MAC_PROP_WL_CAPABILITY:
		wl_get_capability(sc, wldp_buf);
		break;
	case MAC_PROP_WL_WPA:
		wl_get_wpa(sc, wldp_buf);
		break;
	case MAC_PROP_WL_SCANRESULTS:
		wl_get_scanresults(sc, wldp_buf);
		break;
	case MAC_PROP_WL_KEY_TAB:
	case MAC_PROP_WL_KEY:
	case MAC_PROP_WL_DELKEY:
	case MAC_PROP_WL_SETOPTIE:
	case MAC_PROP_WL_MLME:
		cmn_err(CE_WARN, "ieee80211_setprop: opmode %d err\n",
						(int)wldp_pr_num);
		err = EINVAL;
		break;
	default:
		cmn_err(CE_WARN, "ieee80211_setprop: opmode %d not supported\n",
						(int)wldp_pr_num);
		err = ENOTSUP;
		break;
	}
	NDIS_UNLOCK(sc);

	return (err);
}
