/*
 *            Make a Windows wifi driver linkable
 *
 * Copyright (c) 2013 Jean-Pierre Andre
 *
 *     - create symbolic references to external functions
 *     - find absolute references to data and code and make them relocatable
 *     - find relative references to variables and make them relocatable
 *     - insert calls to register dump for debugging
 *     - also patching multimedia instructions referencing xmm registers
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the NTFS-3G
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define ALONE 0 /* set to 1 for testing */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#if ALONE
void exit(int);
#else
#include "cdefs.h"
#include "pe_var.h"
#endif

               /* select a syntax for the output */
#ifdef SOZOBON
#define INTELSYNTAX 1
#else
#define INTELSYNTAX 0
#endif

#if ALONE
#define INFILE64 "bcmwl564.newsys"
#define INFILE32 "bcmwl5.sys"
#endif

#ifdef FALSE
typedef int boolean;
#else
typedef enum { FALSE, TRUE } boolean;
#endif

enum { RNONE, RRELOC, RPATCH };

typedef enum { RAX, RCX, RDX, RBX, RSP, RBP, RSI, RDI,
               R8, R9, R10, R11, R12, R13, R14, R15,
               XMM0, XMM1, XMM2, XMM3, XMM4, XMM5, XMM6, XMM7,
               RIP } REGS;

struct PATCH
   {
   struct PATCH *next;
   enum { TPATCHSYST, TPATCHR, TXMM, TTRACE, TINSERT, TCALL,
                       TRIP, TVAR, TIMPORT } type;
   enum { PDOUBLE = 1, PNEXT = 2 } flags;
   long loc;
   const char *name;
   const char *back;
   const char *code;
   int size;
   int nextsize;
   int offset;
   char srcreg;
   char dstreg;
   char xreg;
   char xshift;
   char data[32];
   } ;

const char *regs[] = /* must match the enum REGS */
   {
   "rax", "rcx", "rdx", "rbx", "rsp", "rbp", "rsi", "rdi",
   "r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15",
   "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7",
   "rip"
   } ;

boolean x8664;
char *relocs;
const char *img;
long filesize;
struct PATCH *firstpatch;
long patchoffset;
int tracenum;
int callnum;
boolean verbose = FALSE;

unsigned long loccode, rvacode, sizecode;
unsigned long locdata;
unsigned long locvars, rvavars, sizevars;
unsigned long locreloc, sizereloc, locimports;
int year;
#if ALONE
unsigned long locinit, sizeinit;
#endif

void defreloc(long loc)
   {
   char *p;

   p = &relocs[loc/4];
   *p |= RRELOC << ((loc & 3) << 1);
   }

struct PATCH *defpatch(long loc, const char *name, int size, int type)
   {
   struct PATCH *patch;
   char *copy;
   int shf;
   char *p;

   patch = (struct PATCH*)malloc(sizeof(struct PATCH));
   copy = (char*)malloc(strlen(name) + 1);
   if (patch && copy)
      {
      strcpy(copy, name);
      patch->next = firstpatch;
      patch->type = type;
      patch->loc = loc;
      patch->name = copy;
      patch->size = size;
      patch->back = (const char*)NULL;
      patch->flags = 0;
      firstpatch = patch;
      p = &relocs[loc/4];
      shf = (loc & 3) << 1;
      *p = (*p & ~(3 << shf)) | (RPATCH << shf);
      }
   return (patch);
   }

struct PATCH *findpatch(long loc)
   {
   struct PATCH *p;

   p = firstpatch;
   while (p && (p->loc != loc))
      p = p->next;
   return (p);
   }

boolean deftrace(long rva)
   {
   char name[16];
   char *back;
   boolean err;
   struct PATCH *patch;

   err = FALSE;
   tracenum++;
   sprintf(name,"trace%d",tracenum);
   patch = defpatch(rva + loccode - rvacode, name, 0, TTRACE);
   if (patch)
      {
      back = (char*)malloc(16);
      if (back)
         {
         sprintf(back,"trback%d",tracenum);
         patch->back = back;
         }
      }
   if (!patch || !back)
      {
      fprintf(stderr,"Could not insert trace\n");
      err = TRUE;
      }
   return (err);
   }

boolean defcall(long rva, const char *name)
   {
   boolean err;
   struct PATCH *patch;

   err = FALSE;
   callnum++;
   patch = defpatch(rva + loccode - rvacode, name, 0, TCALL);
   if (!patch)
      {
      fprintf(stderr,"Could not insert call\n");
      err = TRUE;
      }
   return (err);
   }

boolean definsert(long rva, const char *name, const char *code)
   {
   boolean err;
   struct PATCH *patch;

   err = FALSE;
   callnum++;
   patch = defpatch(rva + loccode - rvacode, name, 0, TINSERT);
   if (patch)
      patch->code = code;
   else
      {
      fprintf(stderr,"Could not insert code at RVA 0x%lx\n",(long)rva);
      err = TRUE;
      }
   return (err);
   }

boolean findrelocs(void)
   {
   boolean err;
   boolean done;
   const char *ptr;
   unsigned long base;
   unsigned long loc;
   int size;
   int offset;
   int reloccnt;
   int c;
   int i;

   err = FALSE;
   relocs = (char*)malloc(filesize/4);
   if (relocs)
      {
      for (i=0; 4*i<filesize; i++)
         relocs[i] = 0;
      reloccnt = 0;
      ptr = &img[locreloc];
      loc = locreloc;
      do {
         base = *ptr++ & 255;
         for (i=8; i<32; i+=8)
            base += (*ptr++ & 255) << i;
         size = *ptr++ & 255;
         for (i=8; i<32; i+=8)
            size += (*ptr++ & 255) << i;
         loc += 8;
         if (verbose)
            fprintf(stderr,"Reloc base 0x%lx size 0x%lx\n",
                             (long)base,(long)size);
         if ((size && (size <= 8)) || (size & 3) || (base & 4095))
            {
            fprintf(stderr,"Bad relocation block\n");
            err = TRUE;
            }
         done = !size;
         size -= 8;
         while (!err && (size > 0))
            {
            offset = *ptr++ & 255;
            c = *ptr++ & 255;
            loc += 2;
            if (offset || c)
               {
               if ((c & 0xf0) != (x8664 ? 0xa0 : 0x30))
                  {
                  fprintf(stderr,"bad relocation mode\n");
                  err = TRUE;
                  }
               else
                  {
                  offset += (c & 15) << 8;
                  if (((base + offset) < loccode)
                    || ((base + offset) >= locreloc))
                     {
                     fprintf(stderr,"Bad relocation (loc 0x%lx)\n",
                                    base + offset);
                     err = TRUE;
                     }
                  else
                     {
                     defreloc(base + offset);
                     reloccnt++;
                     }
                  }
               }
            size -= 2;
            }
         } while (!err && !done && (loc < (locreloc + sizereloc)));
      fprintf(stderr,"%d %s relocations\n",
               reloccnt, (x8664 ? "quadword" : "doubleword"));
      }
   else
      {
      fprintf(stderr,"Could not allocate relocs\n");
      err = TRUE;
      }
   return (err);
   }

boolean findimports(void)
   {
   char buf[256];
   const char *ptr;
   char *p;
   int ptrsz;
   int num;
   int impcnt;
   boolean err;
   boolean ndis5;
   int i;
   int dllnum;
   unsigned long namesloc;
   unsigned long ptrsloc;
   unsigned long dllnameloc;
   unsigned long nameloc;

   ptrsz = (x8664 ? 8 : 4);
   err = FALSE;
   ndis5 = FALSE;
   dllnum = 0;
   impcnt = 0;
   do {
           /* get an import block (20 bytes) */
      ptr = &img[locimports + 20*dllnum];
      namesloc = *ptr++ & 255;
      for (i=8; i<32; i+=8)
         namesloc += (*ptr++ & 255) << i;
      for (i=0; i<8; i++)
         *ptr++;
      dllnameloc = *ptr++ & 255;
      for (i=8; i<32; i+=8)
         dllnameloc += (*ptr++ & 255) << i;
      ptrsloc = *ptr++ & 255;
      for (i=8; i<32; i+=8)
          ptrsloc += (*ptr++ & 255) << i;
//fprintf(stderr,"namesloc 0x%lx ptrsloc 0x%lx\n",(long)namesloc,(long)ptrsloc);
      if (dllnameloc)
         {
         num = 0;
         do {
            ptr = &img[namesloc + ptrsz*num];
            nameloc = *ptr++ & 255;
            for (i=8; i<32; i+=8)
               nameloc += (*ptr++ & 255) << i;
            if (nameloc)
               {
               ptr = &img[nameloc + 2];
               p = buf;
               do {
                  *p = *ptr++;
                  } while (*p++);
               if (verbose)
                  fprintf(stderr,"Import 0x%lx %s\n",ptrsloc + ptrsz*num, buf);
               defpatch(ptrsloc + 8*num, buf, 8, TIMPORT);
                  /* check obsoleted functions which were needed for NDIS5 */
               if (!strcmp(buf,"NdisAllocateBuffer")
                 || !strcmp(buf,"NdisReadPciSlotInformation"))
                  ndis5 = TRUE;
               num++;
               impcnt++;
               }
            } while (nameloc);
         }
      dllnum++;
      } while (!err && dllnameloc);
   if (!ndis5)
      fprintf(stderr,"*** Warning : this driver does not appear to comply with NDIS5\n");
   fprintf(stderr,"%d imported functions\n",impcnt);
   return (err);
   }

boolean settrace(void)
   {
   tracenum = 0;

   switch (year)
      {
      case 2007 :
/* tests
   deftrace(0x34b1f);
   deftrace(0x34b2f);
   deftrace(0x34e20);
   deftrace(0x34bb0);
*/
         break;
      case 2010 :
         break;
      default :
         break;
      }
   if (tracenum)
      fprintf(stderr,"%d %s inserted\n",
                 tracenum, (tracenum > 1 ? "traces" : "trace"));
   return (FALSE);
   }

boolean setcalls(void)
   {
   callnum = 0;

   switch (year)
      {
      case 2007 :
         break;
      case 2008 :
               /* Fix conditional branch to relocated code */
         definsert(0x45605,"reloc45605",
#if INTELSYNTAX
                           " cmp byte ptr 1[rbx],3\n jnc drv_data+218635\n"
                           " lea rdx,96[rsp]\n jmp drv_data+218692\n");
#else
                           " cmpb $3,1(%rbx)\n jnc drv_data+218635\n"
                           " leaq 96(%rsp),%rdx\n jmp drv_data+218692\n");
#endif
         break;
      case 2010 :
         break;
      default :
         break;
      }
   if (callnum)
      fprintf(stderr,"%d %s inserted\n",
                 callnum, (callnum > 1 ? "patches" : "patch"));
   return (FALSE);
   }

/*
 *                    Determine the size of an instruction
 *
 *         This is needed for patching an instruction
 */

enum
   {
   PREXB =        1,
   PREXX =        2,
   PREXR =        4,
   PREXW =        8,
   PREX =      0x40,
   P8086 =     0x80,
   PSEGOVER = 0x100,
   PADOVER =  0x200,
   PBDATA =   0x400,
   PWDATA =   0x800,
   PQDATA =  0x1000, // == PREXW ?
   PTDATA =  0x2000,
   PODATA =  0x4000,
   PNDATA =  0x8000,
   PREP =   0x10000,
   PREPN =  0x20000,
   PCS =    0x40000,
   PDS =    0x80000,
   PES =   0x100000,
   PFS =   0x200000,
   PGS =   0x400000,
   PSS =   0x800000,
   };

int rmsize(const char *p, unsigned long prefixes)
   {
   int rm;
   int sib;
   int size;

   rm = *p & 255;
   if (prefixes & PADOVER)
      switch (rm & 0xc0)
         {
         case 0x00 : /* indirect or direct */
	    if ((rm & 7) == 6) size = 3;
            else size = 1;
            break;
         case 0x40 : /* short offset */
            size = 2;
            break;
         case 0x80 : /* long offset */
            size = 3;
            break;
         default : /* not allowed here */
            size = 0;
            break;
         }
   else
      if ((rm & 7) == 4)
         {
         sib = p[1] & 255;
         switch (rm & 0xc0)
            {
            case 0x00 : /* direct or indirect */
               if ((sib & 7) == 5) size = 6;
               else size = 2;
               break;
            case 0x40 : /* short offset */
               size = 3;
               break;
            case 0x80 : /* long offset */
               size = 6;
               break;
            default : /* not allowed here */
               size = 0;
               break;
            }
         }
      else
         switch (rm & 0xc0)
            {
            case 0x00 : /* indirect or direct */
               if ((rm & 7) == 5) size = 5;
               else size = 1;
               break;
            case 0x40 : /* short offset */
               size = 2;
               break;
            case 0x80 : /* long offset */
               size = 5;
               break;
            default : /* not allowed here */
               size = 0;
               break;
            }
   return (size);
   }

int instrsize(const char *p, unsigned long prefixes)
   {
   int op;
   int op2;
   int rm;
   int size;

   op = *p & 255;
   op2 = 256;
   size = 0;
   switch (op)
      {
      case 0x0f :
         op2 = p[1] & 255;
         switch (op2)
            {
            case 0x00 :
               rm = p[2] & 255;
               switch (rm & 0x38)
                  {
                  case 0x08 : /* str */
                     if ((rm & 0xc0) == 0xc0) size = 0;
                     else size = 2 + rmsize(&p[2], prefixes);
                     break;
                  }
               break;
            case 0x05 : /* syscall */
            case 0x31 : /* rdtsc */
            case 0x33 : /* rdpmc */
            case 0x34 : /* sysenter */
            case 0x35 : /* sysexit */
               size = 2;
               break;
            case 0x0d :
               rm = p[2] & 255;
               switch (rm & 0x38)
                  {
                  case 0x00 : /* prefetch */
                  case 0x08 : /* prefetchw */
                     size = 2 + rmsize(&p[2], prefixes);
                     break;
                  default :
                     size = 0;
                     break;
                  }
               break;
            case 0x18 :
               rm = p[2] & 255;
               switch (rm & 0x38)
                  {
                  case 0x00 : /* prefetchnta */
                  case 0x08 : /* prefetcht0 */
                  case 0x10 : /* prefetcht1 */
                  case 0x18 : /* prefetcht2 */
                     size = 2 + rmsize(&p[2], prefixes);
                     break;
                  default :
                     size = 0;
                     break;
                  }
               break;
            case 0x20 : /* mov from cr */
            case 0x21 : /* mov from dr */
            case 0x22 : /* mov to cr */
            case 0x23 : /* mov to dr */
               size = 3;
               break;
            case 0x80 :
            case 0x81 :
            case 0x82 : /* jc */
            case 0x83 : /* jnc */
            case 0x84 : /* je */
            case 0x85 : /* jne */
            case 0x86 :
            case 0x87 :
            case 0x88 :
            case 0x89 :
            case 0x8a :
            case 0x8b :
            case 0x8c :
            case 0x8d :
            case 0x8e :
            case 0x8f :
               size = (prefixes & PWDATA ? 4 : 6);
               break;
            case 0x03 : /* lsl */
            case 0x10 : /* ? */
            case 0x11 : /* movups */
            case 0x12 : /* movlps */
            case 0x13 : /* movlps */
            case 0x14 : /* unpcklps */
            case 0x15 : /* unpckhps */
            case 0x1f : /* nop (with rm) */
            case 0x28 : /* movaps */
            case 0x29 : /* movaps */
            case 0x2a : /* cvtpi2ps */
            case 0x2c : /* cvttps2pi */
            case 0x2d : /* cvtps2pi */
            case 0x2e : /* ucomiss */
            case 0x2f : /* comiss */
            case 0x40 :
            case 0x41 :
            case 0x42 :
            case 0x43 :
            case 0x44 : /* cmove */
            case 0x45 : /* cmovne */
            case 0x46 :
            case 0x47 :
            case 0x48 :
            case 0x49 :
            case 0x4a :
            case 0x4b :
            case 0x4c :
            case 0x4d :
            case 0x4e :
            case 0x4f : /* cmovg */
            case 0x50 : /* movmskps (mmx) */
            case 0x51 : /* sqrtps */
            case 0x54 : /* andps */
            case 0x55 : /* andnps */
            case 0x56 : /* orps */
            case 0x57 : /* xorps */
            case 0x58 : /* addps */
            case 0x59 : /* mulps */
            case 0x5a : /* cvtps2pd */
            case 0x5b : /* cvtdq2ps */
            case 0x5c : /* subps */
            case 0x5e : /* divps */
            case 0x6e : /* movd (mmx) */
            case 0x6f : /* movq */
            case 0x70 : /* pshufw */
            case 0x74 : /* pcmpeqb */
            case 0x75 : /* pcmpeqw */
            case 0x76 : /* pcmpeqd */
            case 0x7e : /* movd (mmx) */
            case 0x7f : /* movq */
            case 0x90 :
            case 0x91 :
            case 0x92 : /* setc */
            case 0x93 : /* setnc */
            case 0x94 : /* sete */
            case 0x95 : /* setne */
            case 0x96 :
            case 0x97 :
            case 0x98 :
            case 0x99 :
            case 0x9a :
            case 0x9b :
            case 0x9c :
            case 0x9d :
            case 0x9e :
            case 0x9f :
            case 0xa3 : /* bt */
            case 0xa5 : /* shld cl */
            case 0xab : /* bts */
            case 0xad : /* shrd cl */
            case 0xaf : /* imul */
            case 0xb0 : /* cmpxchg */
            case 0xb1 : /* cmpxchg */
            case 0xb2 : /* lss */
            case 0xb3 : /* btr */
            case 0xb4 : /* lfs */
            case 0xb5 : /* lgs */
            case 0xb6 : /* movzx */
            case 0xb7 : /* movzx */
            case 0xbb : /* btc */
            case 0xbc : /* bsf */
            case 0xbd : /* bsr */
            case 0xbe : /* movsx */
            case 0xbf : /* movsx */
            case 0xc0 : /* xadd */
            case 0xc1 : /* xadd */
            case 0xc3 : /* movnti */
            case 0xc4 : /* pinsrw */
            case 0xc6 : /* shufps */
            case 0xd4 : /* paddq */
            case 0xd6 : /* mpvq2dq */
            case 0xd7 : /* pmovmskb */
            case 0xdb : /* pand */
            case 0xdf : /* pandn */
            case 0xe6 : /* cvtdq2pd */
            case 0xee : /* pmaxsw */
            case 0xef : /* pxor */
            case 0xd3 : /* psrlq */
            case 0xf3 : /* psllq */
            case 0xf8 : /* psubb */
            case 0xf9 : /* psubw */
            case 0xfa : /* psubd */
            case 0xfb : /* psubq */
            case 0xfc : /* paddb */
            case 0xfd : /* paddw */
            case 0xfe : /* paddd */
               rm = p[2] & 255;
               if ((rm & 0xc0) == 0xc0) size = 3;
               else size = 2 + rmsize(&p[2], prefixes);
               break;
            case 0x73 : /* psllq imm */
            case 0xc2 : /* cmpss */
            case 0xc5 : /* pextrw */
               size = 4;
               break;
            case 0xa4 : /* shld imm */
            case 0xac : /* shrd imm */
            case 0xba : /* bt, bts, btr, btc */
               rm = p[2] & 255;
               if ((rm & 0xc0) == 0xc0) size = 4;
               else size = 3 + rmsize(&p[2], prefixes);
               break;
            case 0xae :
               rm = p[2] & 255;
               switch (rm & 0x38)
                  {
                  case 0x00 : /* fxsave */
                  case 0x08 : /* fxrstor */
                  case 0x10 : /* ldmxcsr */
                  case 0x18 : /* stmxcsr */
                     if ((rm & 0xc0) == 0xc0) size = 3;
                     else size = 2 + rmsize(&p[2], prefixes);
                     break;
                  case 0x28 : /* lfence */
                  case 0x30 : /* mfence */
                  case 0x38 : /* sfence */
                     if ((rm & 0xc7) == 0xc7) size = 3;
                     else size = 0;
                     break;
                  }
               break;
            case 0xc7 :
               rm = p[2] & 255;
               switch (rm & 0x38)
                  {
                  case 0x08 : /* cmpxch8b, cmpxch16b */
                     if ((rm & 0xc0) == 0xc0) size = 0;
                     else size = 2 + rmsize(&p[2], prefixes);
                     break;
                  }
               break;
            case 0xa0 : /* push fs */
            case 0xa1 : /* pop fs */
            case 0xa2 : /* cpuid */
            case 0xa8 : /* push gs */
            case 0xa9 : /* pop gs */
            case 0xc8 : /* bswap eax */
            case 0xc9 : /* bswap ecx */
            case 0xca : /* bswap edx */
            case 0xcb : /* bswap ebx */
            case 0xcc : /* bswap esp */
            case 0xcd : /* bswap ebp */
            case 0xce : /* bswap esi */
            case 0xcf : /* bswap edi */
               size = 2;
               break;
            }
         break;
      case 0x05 : /* add imm long */
      case 0x0d : /* or imm long */
      case 0x15 : /* adc imm long */
      case 0x1d : /* sbb imm long */
      case 0x25 : /* and imm long */
      case 0x2d : /* sub imm long */
      case 0x35 : /* xor imm long */
      case 0x3d : /* cmp imm long */
      case 0xa9 : /* test imm long */
         size = (prefixes & PWDATA ? 3 : 5);
         break;
      case 0x40 :
      case 0x41 :
      case 0x42 :
      case 0x43 :
      case 0x44 :
      case 0x45 :
      case 0x46 :
      case 0x47 :
         if (x8664)
            {
            size = instrsize(&p[1],prefixes | PREX);
            if (size) size++;
            }
         else size = 1;
         break;
      case 0x48 :
      case 0x49 :
      case 0x4a :
      case 0x4b :
      case 0x4c :
      case 0x4d :
      case 0x4e :
      case 0x4f :
         if (x8664)
            {
            size = instrsize(&p[1], prefixes | PREX | PQDATA);
            if (size) size++;
            }
         else size = 1;
         break;
                                    /* fixed size 1 */
      case 0x06 : /* push es */
      case 0x07 : /* pop es */
      case 0x0e : /* push cs */
      case 0x16 : /* push ss */
      case 0x17 : /* pop ss */
      case 0x1e : /* push ds */
      case 0x1f : /* pop ds */
      case 0x27 : /* daa */
      case 0x2f : /* das */
      case 0x37 : /* aaa */
      case 0x3f : /* aas */
      case 0x50 : /* push eax */
      case 0x51 : /* push ecx */
      case 0x52 : /* push edx */
      case 0x53 : /* push ebx */
      case 0x54 : /* push esp */
      case 0x55 : /* push ebp */
      case 0x56 : /* push esi */
      case 0x57 : /* push edi */
      case 0x58 : /* pop eax */
      case 0x59 : /* pop ecx */
      case 0x5a : /* pop edx */
      case 0x5b : /* pop ebx */
      case 0x5c : /* pop esp */
      case 0x5d : /* pop ebp */
      case 0x5e : /* pop esi */
      case 0x5f : /* pop edi */
      case 0x60 : /* pusha */
      case 0x61 : /* popa */
      case 0x6c : /* insb */
      case 0x6d : /* insw */
      case 0x6e : /* outsb */
      case 0x6f : /* outsw */
      case 0x90 : /* nop */
      case 0x91 : /* xchg ecx */
      case 0x92 : /* xchg edx */
      case 0x93 : /* xchg ebx */
      case 0x94 : /* xchg esp */
      case 0x95 : /* xchg ebp */
      case 0x96 : /* xchg esi */
      case 0x97 : /* xchg edi */
      case 0x98 : /* cwd */
      case 0x99 : /* cdq */
      case 0x9b : /* wait */
      case 0x9c : /* pushfd */
      case 0x9d : /* popfd */
      case 0x9e : /* sahf */
      case 0x9f : /* lahf */
      case 0xa4 : /* movs */
      case 0xa5 : /* movs */
      case 0xa6 : /* cmps */
      case 0xa7 : /* cmps */
      case 0xaa : /* stos */
      case 0xab : /* stos */
      case 0xac : /* lods */
      case 0xad : /* lods */
      case 0xae : /* scas */
      case 0xaf : /* scas */
      case 0xc3 : /* ret */
      case 0xc9 : /* leave */
      case 0xcb : /* ret */
      case 0xcc : /* int 3 */
      case 0xce : /* into */
      case 0xcf : /* iret */
      case 0xd7 : /* xlat */
      case 0xec : /* in */
      case 0xed : /* in */
      case 0xee : /* out */
      case 0xef : /* out */
      case 0xf0 : /* lock */
      case 0xf4 : /* hlt */
      case 0xf5 : /* cmc */
      case 0xf8 : /* clc */
      case 0xf9 : /* stc */
      case 0xfa : /* cli */
      case 0xfb : /* sti */
      case 0xfc : /* cld */
      case 0xfd : /* std */
         size = 1;
         break;
                                    /* fixed size 2 */
      case 0x04 : /* add imm byte */
      case 0x0c : /* or imm byte */
      case 0x14 : /* adc imm byte */
      case 0x1c : /* sbb imm byte */
      case 0x24 : /* and imm byte */
      case 0x2c : /* sub imm byte */
      case 0x34 : /* xor imm byte */
      case 0x3c : /* cmp imm byte */
      case 0xa8 : /* test imm byte */
      case 0x6a : /* push */
      case 0xb0 : /* mov imm al */
      case 0xb1 : /* mov imm cl */
      case 0xb2 : /* mov imm dl */
      case 0xb3 : /* mov imm bl */
      case 0xb4 : /* mov imm ah */
      case 0xb5 : /* mov imm ch */
      case 0xb6 : /* mov imm dh */
      case 0xb7 : /* mov imm bh */
      case 0xcd : /* int */
      case 0xd4 : /* aam */
      case 0xd5 : /* aad */
      case 0xe4 : /* in */
      case 0xe5 : /* in */
      case 0xe6 : /* out */
      case 0xe7 : /* out */
         size = 2;
         break;
      case 0x63 : /* movsx or arpl */
         if (x8664 && (prefixes & (PREXW | PQDATA)))
            {
            rm = p[1] & 255;
            if ((rm & 0xc0) == 0xc0) size = 2;
            else size = 1 + rmsize(&p[1], prefixes);
            }
         else size = 0;
         break;
      case 0x26 : /* es: */
      case 0x2e : /* cs: */
      case 0x36 : /* ss: */
      case 0x3e : /* ds: */
      case 0x64 : /* fs: */
      case 0x65 : /* gs: */
         size = instrsize(&p[1], prefixes | PSEGOVER);
            /* improve : segment for memory addressing only */
         if (size > 1) size++;
         else
            {
            op2 = p[1] & 255;
            if (((op2 >= 0xa4) && (op2 <= 0xa7))
              || ((op2 >= 0xaa) && (op2 <= 0xaf))) size++;
            else size = 0;
            }
         break;
      case 0x66 :
         size = instrsize(&p[1], prefixes | PWDATA);
         if (size) size++;
         break;
      case 0x67 :
         size = instrsize(&p[1], prefixes | PADOVER);
         if (size) size++;
         break;
      case 0x68 : /* push */
         size = (prefixes & PWDATA ? 3 : 5);
         break;
      case 0x70 :
      case 0x71 :
      case 0x72 : /* jc */
      case 0x73 : /* jnc */
      case 0x74 : /* je */
      case 0x75 : /* jne */
      case 0x76 :
      case 0x77 :
      case 0x78 :
      case 0x79 :
      case 0x7a :
      case 0x7b :
      case 0x7c :
      case 0x7d :
      case 0x7e :
      case 0x7f :
      case 0xe0 : /* loopne */
      case 0xe1 : /* loope */
      case 0xe2 : /* loop */
      case 0xe3 : /* jcxz */
      case 0xeb : /* jmp */
         size = 2;
         break;
      case 0x69 : /* mul imm long, etc */
      case 0x81 : /* add imm long, etc */
      case 0xc7 : /* mov */
         rm = p[1] & 255;
         if ((rm & 0xc0) == 0xc0) size = 2;
         else size = 1 + rmsize(&p[1], prefixes);
         size += (prefixes & PWDATA ? 2 : 4);
         break;
      case 0x6b : /* mul imm short */
      case 0x80 : /* add imm short, etc */
      case 0x82 : /* add imm short, etc */
      case 0x83 : /* add imm short, etc */
      case 0xc6 : /* mov */
      case 0xc0 : /* shift + count */
      case 0xc1 : /* shift + count */
         rm = p[1] & 255;
         if ((rm & 0xc0) == 0xc0) size = 3;
         else size = 2 + rmsize(&p[1], prefixes);
         break;
                                    /* single byte + rm (+ sib) + offset */
      case 0x00 : /* add */
      case 0x01 : /* add */
      case 0x02 : /* add */
      case 0x03 : /* add */
      case 0x08 : /* or */
      case 0x09 : /* or */
      case 0x0a : /* or */
      case 0x0b : /* or */
      case 0x10 : /* adc */
      case 0x11 : /* adc */
      case 0x12 : /* adc */
      case 0x13 : /* adc */
      case 0x18 : /* sbb */
      case 0x19 : /* sbb */
      case 0x1a : /* sbb */
      case 0x1b : /* sbb */
      case 0x20 : /* and */
      case 0x21 : /* and */
      case 0x22 : /* and */
      case 0x23 : /* and */
      case 0x28 : /* sub */
      case 0x29 : /* sub */
      case 0x2a : /* sub */
      case 0x2b : /* sub */
      case 0x30 : /* xor */
      case 0x31 : /* xor */
      case 0x32 : /* xor */
      case 0x33 : /* xor */
      case 0x38 : /* cmp */
      case 0x39 : /* cmp */
      case 0x3a : /* cmp */
      case 0x3b : /* cmp */
      case 0x62 : /* bound */
      case 0x84 : /* test */
      case 0x85 : /* test */
      case 0x86 : /* xchg */
      case 0x87 : /* xchg */
      case 0x88 : /* mov */
      case 0x89 : /* mov */
      case 0x8a : /* mov */
      case 0x8b : /* mov */
      case 0x8c : /* mov seg */
      case 0x8e : /* mov seg */
      case 0xd0 : /* shift short 1 */
      case 0xd1 : /* shift long 1 */
      case 0xd2 : /* shift short cl */
      case 0xd3 : /* shift long cl */
         rm = p[1] & 255;
         if ((rm & 0xc0) == 0xc0) size = 2;
         else size = 1 + rmsize(&p[1], prefixes);
         break;
      case 0xf6 : /* neg, not, test,... */
      case 0xf7 : /* neg, not, test,... */
         rm = p[1] & 255;
         switch (rm & 0x38)
            {
            case 0x00 : /* test */
               size = (op & 1 ? (prefixes & PWDATA ? 2 : 4) : 1);
               if ((rm & 0xc0) == 0xc0) size += 2;
               else size += 1 + rmsize(&p[1], prefixes);
               break;
            case 0x10 : /* not */
            case 0x18 : /* neg */
            case 0x20 : /* mul */
            case 0x28 : /* imul */
            case 0x30 : /* div */
            case 0x38 : /* idiv */
               if ((rm & 0xc0) == 0xc0) size = 2;
               else size = 1 + rmsize(&p[1], prefixes);
               break;
            }
         break;
      case 0x8d : /* lea */
      case 0xc4 : /* les */
      case 0xc5 : /* lds */
         rm = p[1] & 255;
         if ((rm & 0xc0) == 0xc0) size = 0;
         else size = 1 + rmsize(&p[1], prefixes);
         break;
      case 0x8f :
         rm = p[1] & 255;
         switch (rm & 0x38)
            {
            case 0x00 : /* pop */
               if ((rm & 0xc0) == 0xc0) size = 2;
               else size = 1 + rmsize(&p[1], prefixes);
               break;
            default :
               size = 0;
               break;
            }
         break;
      case 0x9a : /* call far */
      case 0xea : /* jmp far */
         size = (prefixes & PADOVER ? 5 : 7);
         break;
      case 0xa0 : /* mov al */
      case 0xa1 : /* mov eax */
      case 0xa2 : /* mov ,al */
      case 0xa3 : /* mov ,eax */
         size = (prefixes & PQDATA ? 9 : (prefixes & PADOVER ? 3 : 5));
         break;
      case 0xb8 : /* mov eax */
      case 0xb9 : /* mov ecx */
      case 0xba : /* mov edx */
      case 0xbb : /* mov ebx */
      case 0xbc :
      case 0xbd :
      case 0xbe : /* mov esi */
      case 0xbf : /* mov edi */
         size = (prefixes & PQDATA ? 9 : (prefixes & PWDATA ? 3 : 5));
         break;
                                    /* fixed size 3 */
      case 0xc2 : /* ret */
      case 0xca : /* ret */
         size = 3;
         break;
                                    /* fixed size 4 */
      case 0xc8 : /* enter */
         size = 4;
         break;
      case 0xd8 : /* float x87 */
      case 0xd9 : /* float x87 */
      case 0xda : /* float x87 */
      case 0xdb : /* float x87 */
      case 0xdc : /* float x87 */
      case 0xdd : /* float x87 */
      case 0xde : /* float x87 */
      case 0xdf : /* float x87 */
         rm = p[1] & 255;
         if (rm < 0xc0)
            size = 1 + rmsize(&p[1], prefixes);
         else size = 2;
         break;
      case 0xe8 : /* call */
      case 0xe9 : /* jmp */
         size = (prefixes & (PADOVER | P8086) ? 3 : 5);
         break;
      case 0xf2 : /* repne */
         size = instrsize(&p[1], prefixes | PREPN);
         if (size) size++;
         break;
      case 0xf3 : /* rep */
         size = instrsize(&p[1], prefixes | PREP);
         if (size) size++;
         break;
      case 0xfe :
         rm = p[1] & 255;
         switch (rm & 0x38)
            {
            case 0x00 : /* inc short */
            case 0x08 : /* dec short */
               if ((rm & 0xc0) == 0xc0) size = 2;
               else size = 1 + rmsize(&p[1], prefixes);
               break;
            default :
               size = 0;
               break;
            }
         break;
      case 0xff :
         rm = p[1] & 255;
         switch (rm & 0x38)
            {
            case 0x00 : /* inc long */
            case 0x08 : /* dec long */
            case 0x10 : /* call */
            case 0x18 : /* call far */
            case 0x20 : /* jmp */
            case 0x28 : /* jmp far */
            case 0x30 : /* push */
               if ((rm & 0xc0) == 0xc0) size = 2;
               else size = 1 + rmsize(&p[1], prefixes);
               break;
            default :
               size = 0;
               break;
            }
         break;
      default :
         size = 0;
         break;
      }
   return (size);
   }

/*
 *                Determine whether the address is relative to ip
 *         Returns the location of the last byte of the relative address.
 *         (largely incomplete, enough for bcmwl564.sys)
 */

int relative(const char *p)
   {
   int rip;

   rip = 0;
   switch (*p & 255)
      {
      case 0x0f :
         if ((p[1] & 0xf0) == 0x80) rip = 5;
         break;
      case 0x40 :
      case 0x41 :
      case 0x42 :
      case 0x43 :
      case 0x44 :
      case 0x45 :
      case 0x46 :
      case 0x47 :
      case 0x48 :
      case 0x49 :
      case 0x4a :
      case 0x4b :
      case 0x4c :
      case 0x4d :
      case 0x4e :
      case 0x4f :
         switch (p[1] & 255)
            {
            case 0x38 :
            case 0x39 :
            case 0x83 :
            case 0x89 :
            case 0x8b :
            case 0x8d :
               if ((p[2] & 0xc7) == 0x05) rip = 6;
               break;
            case 0xff : /* the prefix is meaningless for call/jmp */
               switch (p[2] & 255)
                  {
                  case 0x15 :
                  case 0x25 :
                     rip = 6;
                     break;
                  }
               break;
            }
         break;
      case 0x70 :
      case 0x71 :
      case 0x72 :
      case 0x73 :
      case 0x74 :
      case 0x75 :
      case 0x76 :
      case 0x77 :
      case 0x78 :
      case 0x79 :
      case 0x7a :
      case 0x7b :
      case 0x7c :
      case 0x7d :
      case 0x7e :
      case 0x7f :
      case 0xeb :
         rip = 1;
         break;
      case 0x38 :
      case 0x39 :
      case 0x83 :
      case 0x89 :
      case 0x8b :
      case 0x8d :
         if ((p[1] & 0xc7) == 0x05) rip = 5;
         break;
      case 0xe8 :
      case 0xe9 :
         rip = 4;
         break;
      case 0xff :
         switch (p[1] & 255)
            {
            case 0x15 :
            case 0x25 :
               rip = 5;
               break;
            }
         break;
      }
   return (rip);
   }

/*
 *                Determine whether the address is an absolute one
 *         followed by an immediate value.
 *         Absolute addresses which appear as an immediate value are
 *         not considered here.
 *         Returns the location of the last byte of the absolute address.
 */

int absolimmed(const char *p)
   {
   int abs;
   int q;

   abs = 0;
   if (*p == 0x66)
      abs++;
   if (x8664 && ((p[abs] & 0xf0) == 0x40))
      abs++;
   switch (p[abs] & 255)
      {
      case 0x80 : /* add, sub, etc. */
      case 0x81 : /* add, sub, etc. */
      case 0x82 : /* add, sub, etc. */
      case 0x83 : /* add, sub, etc. */
      case 0xc6 : /* mov */
      case 0xc7 : /* mov */
      case 0xf6 : /* test */
      case 0xf7 : /* test */
         abs++;
         q = p[abs] & 0xc7;
         if (q == 0x05)
            abs += 4;      /* direct */
         else
            if (q == 0x84)
               abs += 5;   /* indexed */
            else
               if ((q & 0xc0) == 0x80)
                  abs += 4;  /* based */
               else
                  abs = 0;  /* other */
         break;
      default :
         abs = 0;
         break;
      }
   if (x8664 && abs)
      abs += 4;
   return (abs);
   }

boolean ismovdqu(const char *p)
   {
   boolean movdqu;

   movdqu = ((*p & 255) == 0xf3)
          && (p[1] == 0x0f)
          && ((p[2] == 0x6f) || (p[2] == 0x7f));
   if (!movdqu)
      movdqu = ((*p & 255) == 0xf3)
             && (p[1] >= 0x40) && (p[1] <= 0x4f)
             && (p[2] == 0x0f)
             && ((p[3] == 0x6f) || (p[3] == 0x7f));
   return (movdqu);
   }

boolean ismovaps(const char *p)
   {
   boolean movaps;

   movaps = (p[0] == 0x0f)
          && ((p[1] == 0x28) || (p[1] == 0x29));
   if (!movaps)
      movaps = (p[0] >= 0x40) && (p[0] <= 0x4f)
             && (p[1] == 0x0f)
             && ((p[2] == 0x28) || (p[2] == 0x29));
   return (movaps);
   }

/*
 *        Get source and destination of a movdqu or movaps
 *        Largely incomplete...
 */

boolean splitxmm(struct PATCH *patch, const char *buf)
   {
   boolean err;
   const char *p;
   int pref;
   REGS rr;
   REGS rb;
   REGS rx;
   int sx;

   err = FALSE;
   p = buf;
         /* skip prefix, assuming this is either movdqu of movaps */
   if ((*p & 255) == 0xf3)
      p++;
   if ((p[0] >= 0x40) && (p[0] <= 0x4f))
      pref = *p++;
   else pref = 0;
   if ((p[0] == 0x0f)
          && ((p[1] == 0x6f) || (p[1] == 0x7f)
            || (p[1] == 0x28) || (p[1] == 0x29)))
      {
      rx = RSP;
      sx = 0;
      rr = ((p[2] >> 3) & 7) + XMM0;
      if ((p[2] & 0xc7) == 5) /* RIP */
         {
         rb = RIP;
         patch->offset = ((((((p[6] & 255) << 8)
                            + (p[5] & 255)) << 8)
                            + (p[4] & 255)) << 8)
                            + (p[3] & 255);
         }
      else
         if ((p[2] & 7) == 4) /* with SIB */
            {
            rb = (p[3] & 7) + (pref & 1 ? R8 : RAX);
            rx = ((p[3] >> 3) & 7) + RAX;
            sx = 1 << ((p[3] >> 6) & 3);
            if ((p[2] & 0xc0) == 0x80)
               patch->offset = ((((((p[7] & 255) << 8)
                                  + (p[6] & 255)) << 8)
                                  + (p[5] & 255)) << 8)
                                  + (p[4] & 255);
            else
               if ((p[2] & 0xc0) == 0x40)
                  patch->offset = p[4];
               else
                  if ((p[2] & 0xc0) == 0x00)
                     patch->offset = 0;
            }
         else /* no SIB */
            {
            rb = (p[2] & 7) + (pref & 1 ? R8 : RAX);
            if ((p[2] & 0xc0) == 0x80)
               patch->offset = ((((((p[6] & 255) << 8)
                                  + (p[5] & 255)) << 8)
                                  + (p[4] & 255)) << 8)
                                  + (p[3] & 255);
            else
               if ((p[2] & 0xc0) == 0x40)
                  patch->offset = p[3];
               else
                  patch->offset = 0;
            }
      if ((p[1] == 0x6f) || (p[1] == 0x28)) /* move to xmm */
         {
         patch->dstreg = rr;
         patch->srcreg = rb;
         }
      else /* move from xmm */
         {
         patch->srcreg = rr;
         patch->dstreg = rb;
         }
      patch->xreg = rx;
      patch->xshift = sx;
/*
fprintf(stderr,"srcreg %d %s dstreg %d %s xreg %s*%d offset %d\n",
patch->srcreg,regs[patch->srcreg],patch->dstreg,regs[patch->dstreg],
regs[patch->xreg],patch->xshift,patch->offset);
*/
      }
   return (err);
   }

/*
 *               Find changes based on analyzing the code
 *         - access to variables
 *         - use of xmm registers
 *         - access to system data at 0xfffff780xxxxxxxx
 */

boolean findchanges(void)
   {
   char buf[16];
   char name[16];
#if ALONE
#else
   image_section_header sect_hdr;
#endif
   int sectioncnt;
   boolean err;
   const char *ptr;
   struct PATCH *patch;
   int section;
   long locsection;
   long endsection;
   long high;
   char *back;
   long loc;
   long offs;
   unsigned long rva;
   boolean var;
   int xmmnum;
   int xmmcnt;
   int varcnt;
   int syscnt;
   int cnt;
   int rip;
   int sz;
   int sz2;
   int i;

   err = FALSE;
   xmmnum = 0;
   xmmcnt = 0;
   varcnt = 0;
   syscnt = 0;
#if ALONE
   sectioncnt = 2;
#else
   sectioncnt = pe_numsections((vm_offset_t)img);
#endif
   for (section=0; (section<sectioncnt) && !err; section++)
      {
#if ALONE
      if (section)
         {
         locsection = locinit;
         endsection = locinit + sizeinit;
         }
      else
         {
         locsection = loccode;
         endsection = loccode + sizecode;
         }
#else
      err = pe_get_section_header((vm_offset_t)img, section, &sect_hdr);
      locsection = sect_hdr.ish_rawdataaddr;
      endsection = locsection + sect_hdr.ish_rawdatasize;
         /* Only check sections containing executable code */
      if (!err && (sect_hdr.ish_characteristics & 0x20000000))
#endif
         {
         loc = locsection;
         ptr = &img[loc];
         cnt = 0;
         do {
            while (cnt < 16)
               {
               buf[cnt] = *ptr++ & 255;
               cnt++;
               }
            sz = instrsize(buf,0);
            if (sz)
               {
                        /* check whether this is an access to a variable */
               if (x8664)
                  {
                  rip = relative(buf);
                  if (rip)
                     {
                     offs = ((((((((long)buf[rip]) & 255) << 8)
                               + (((long)buf[rip - 1]) & 255)) << 8)
                               + (((long)buf[rip - 2]) & 255)) << 8)
                               + (((long)buf[rip - 3]) & 255);
                     rva = rvacode - loccode + loc + sz + offs;
                           /*
                            * Here, we assume all variables to be located in
                            * the uninitialized part of .data (BSS)
                            * The variables located in the initialized part
                            * should be copied to the DATA section. Currently
                            * they are kept in the CODE section (read-only).
                            */
                     var = (rva >= rvavars) && (rva < (rvavars + sizevars));
                     if (var)
                        {
                        if (verbose)
                           fprintf(stderr,"At loc 0x%lx rip %d offs 0x%lx rva 0x%lx\n",
                                           loc, rip, offs, (long)rva);
                        defpatch(loc + rip - 3, "rip", 4, TRIP);
                        varcnt++;
                        }
                     }
                  }
               else
                  {
                  int absi;
                  int b;
                  char r;

                  absi = 0;
                  if (sz > 4)
                     {
                        /*
                         * Usually the address is at the end of instr,
                         * including when used as an immediate value.
                         */
                     r = relocs[(loc + sz - 4)/4];
                     b = (r >> (((loc + sz - 4) & 3) << 1)) & 3;
                     if (b == RRELOC)
                        absi = sz - 1;
                     else
                        if (sz > 6)
                           {
                             /*
                              * The address may be followed by an immediate
                              * value.
                              * The case where both the operand and the
                              * immediate value are addresses is not
                              * examined.
                              */
                           absi = absolimmed(buf);
                           if (absi)
                              {
                              r = relocs[(loc + absi - 3)/4];
                              b = (r >> (((loc + absi - 3) & 3) << 1)) & 3;
                              if (b != RRELOC)
                                 absi = 0;
                              }
                           }
                     }
                  if (absi)
                     {
                     rva  = ((((((((long)buf[absi]) & 255) << 8)
                               + (((long)buf[absi - 1]) & 255)) << 8)
                               + (((long)buf[absi - 2]) & 255)) << 8)
                               + (((long)buf[absi - 3]) & 255);
//fprintf(stderr,"rva 0x%lx at loc 0x%lx, rvavars 0x%lx absi %d\n",(long)rva,(long)loc,(long)rvavars,absi);
                     var = (rva >= rvavars) && (rva < (rvavars + sizevars));
                     if (var)
                        {
                        if (verbose)
                           fprintf(stderr,"At loc 0x%lx absi %d rva 0x%lx\n",
                                           loc, absi, (long)rva);
                        defpatch(loc + sz - 4, "var", 4, TVAR);
                        varcnt++;
                        }
                     }
                  }
               if ((loc < endsection) && (ismovdqu(buf) || ismovaps(buf)))
                  {
                  back = (char*)malloc(16);
                  if (back)
                     {
                     xmmcnt++;
                     xmmnum++;
                     sprintf(name,"xmmpatch%d",xmmnum);
                     sprintf(back,"xmmback%d",xmmnum);
                     if (verbose)
                        fprintf(stderr,"At loc 0x%lx movdqu/movaps rva 0x%lX name %s\n",
                                            loc,loc+rvacode-loccode,name);
                     patch = defpatch(loc,name,sz,TXMM);
                     if (patch)
                        {
                        patch->back = back;
                        err = splitxmm(patch,buf);
                        if (sz < 5)
                           {
                                /* difficult situation : not enough space for a jump */
                           sz2 = instrsize(&buf[sz],0);
                           if (ismovdqu(&buf[sz]) || ismovaps(&buf[sz]))
                              {
                                 /* another xmm reference : just mention it */
                              patch->flags |= PDOUBLE;
                              patch->nextsize = sz2;
                              }
                           else
                              {
                                /* otherwise, save next instruction */
                              patch->flags |= PNEXT;
                              patch->nextsize = sz2;
                              for (i=0; i<sz2; i++)
                                 patch->data[i] = buf[sz + i];
                              sz += sz2;
                              }
                           }
                        }
                     else err = TRUE;
                     }
                  }
                        /* check whether this is an access to system data */
               if (x8664
                 && ((buf[0] & 0xf8) == 0x48)
                 && ((buf[1] & 0xf8) == 0xb8))
                  {
                  high = ((((((((long)buf[9]) & 255) << 8)
                            + (((long)buf[8]) & 255)) << 8)
                            + (((long)buf[7]) & 255)) << 8)
                            + (((long)buf[6]) & 255);
                  if (high == 0xfffff780)
                     {
                     syscnt++;
                     sprintf(name,"systemdata%d",syscnt);
                     defpatch(loc + 2, name, 8,TPATCHSYST);
                     if (verbose)
                        fprintf(stderr,"At loc 0x%lx %s\n",loc + 2, name);
                     }
                  }
               }
            else
               {
                 /* data may be found within code, skip one byte and retry */
               if (verbose)
                  {
                  fprintf(stderr,"%06lx %02x *unknown*\n",(long)loc,buf[0] & 255);
                  }
               sz = 1;
               }
            for (i=0; i<16-sz; i++) buf[i] = buf[i + sz];
            loc += sz;
            cnt -= sz;
            } while (sz && (loc < endsection));
         }
      }
   if (!err)
      {
      fprintf(stderr,"%d accesses to variables\n",varcnt);
      fprintf(stderr,"%d accesses to system data\n",syscnt);
      fprintf(stderr,"%d uses of xmm registers\n",xmmcnt);
      }
   return (err);
   }

const char *condjmp(num)
   {
   const char *op;

   switch (num)
      {
      case 0 : op = "jo"; break;
      case 1 : op = "jno"; break;
      case 2 : op = "jc"; break;
      case 3 : op = "jnc"; break;
      case 4 : op = "je"; break;
      case 5 : op = "jne"; break;
      case 6 : op = "jbe"; break;
      case 7 : op = "ja"; break;
      case 8 : op = "js"; break;
      case 9 : op = "jns"; break;
      case 10 : op = "jp"; break;
      case 11 : op = "jnp"; break;
      case 12 : op = "jl"; break;
      case 13 : op = "jge"; break;
      case 14 : op = "jle"; break;
      case 15 : op = "jg"; break;
      default : op = "error"; break;
      }
   return (op);
   }

boolean buildheader(FILE *out)
   {
   long loc;
   const char *imp;
   struct PATCH *patch;

#if INTELSYNTAX
   fprintf(out,"          .x86_64\n");
   fprintf(out,"          .xmm\n");
   fprintf(out,"          public drv_data\n");
   fprintf(out,"          public drv_data_len\n");
   fprintf(out,"          extrn dumpregs:far\n");
#else
   fprintf(out,"          .globl drv_data\n");
   fprintf(out,"          .globl drv_data_len\n");
   fprintf(out,"          .extern dumpregs\n");
#endif
// TODO better scan the patch list
   for (loc=0; loc<filesize; loc++)
      switch ((relocs[loc/4] >> ((loc & 3)*2)) & 3)
         {
         case RPATCH :
            patch = findpatch(loc);
            if (patch)
               switch (patch->type)
                  {
                  case TIMPORT :
                     imp = patch->name;
                        /*
                         * Standard C run-time functions require Windows ABI,
                         * make sure not to call implementations for Unix.
                         * (not needed in 32-bit mode)
                         * Note : the list should be much extended, but
                         * that leads to developing a variant of all the
                         * possible functions in ndisapi, so some generic
                         * procedure has to be established.
                         */
                     if (x8664
                       && (!strcmp(imp,"sprintf")
                         || !strcmp(imp,"strncmp")
                         || !strcmp(imp,"strncpy")
                         || !strcmp(imp,"strchr")
                         || !strcmp(imp,"strrchr")
                         || !strcmp(imp,"memcpy")))
                        {
#if INTELSYNTAX
                        fprintf(out,"          extrn win_%s:near\n",imp);
                        fprintf(out,"%s  equ win_%s\n",imp,imp);
#else
                        fprintf(out,"   .extern win_%s\n",imp);
                        fprintf(out,"   .equ %s, win_%s\n",imp,imp);
#endif
                        }
                     else
#if INTELSYNTAX
                        fprintf(out,"          extrn %s:near\n",imp);
#else
                        fprintf(out,"         .extern %s\n",imp);
#endif
                     break;
                  }
            break;
         default :
            break;
         }

#if INTELSYNTAX
   fprintf(out,"DGROUP    group   _BSS,_DATA\n");
   fprintf(out,"_BSS      segment para public '.bss'\n");
   fprintf(out,"          align 16\n");
   fprintf(out,"variables db %ld dup(?)\n",(long)sizevars);
   fprintf(out,"          align 16\n");
   fprintf(out,"rxmm0l    dq 1 dup(?)\n");
   fprintf(out,"rxmm0h    dq 1 dup(?)\n");
   fprintf(out,"rxmm1l    dq 1 dup(?)\n");
   fprintf(out,"rxmm1h    dq 1 dup(?)\n");
   fprintf(out,"rxmm5l    dq 1 dup(?)\n");
   fprintf(out,"rxmm5h    dq 1 dup(?)\n");
   fprintf(out,"rxmm6l    dq 1 dup(?)\n");
   fprintf(out,"rxmm6h    dq 1 dup(?)\n");
   for (patch=firstpatch; patch; patch=patch->next)
      if (patch->type == TPATCHSYST)
         fprintf(out,"%s dq 2 dup(?)\n",patch->name);
   fprintf(out,"_BSS      ends\n");
   fprintf(out,"_DATA     segment para public '.data'\n");
   fprintf(out,"drv_data_len dd %ld\n",(long)filesize);
   fprintf(out,"_DATA     ends\n");
#else
   fprintf(out,"   .bss\n");
   fprintf(out,"   .align 16\n");
   fprintf(out,"   .comm  variables,%ld,%ld\n",(long)sizevars,(long)sizevars);
   fprintf(out,"   .align 16\n");
   fprintf(out,"   .comm  rxmm0l,8,8\n");
   fprintf(out,"   .comm  rxmm0h,8,8\n");
   fprintf(out,"   .comm  rxmm1l,8,8\n");
   fprintf(out,"   .comm  rxmm1h,8,8\n");
   fprintf(out,"   .comm  rxmm5l,8,8\n");
   fprintf(out,"   .comm  rxmm5h,8,8\n");
   fprintf(out,"   .comm  rxmm6l,8,8\n");
   fprintf(out,"   .comm  rxmm6h,8,8\n");
   for (patch=firstpatch; patch; patch=patch->next)
      if (patch->type == TPATCHSYST)
         fprintf(out,"   .comm %s,16,16\n",patch->name);
   fprintf(out,"   .data\n");
   fprintf(out,"   .align 16\n");
   fprintf(out,"drv_data_len :\n");
   fprintf(out,"   .long %ld\n",(long)filesize);
#endif
   return (FALSE);
   }

/*
 *         Relocate a relative instruction within a patch
 */

int relocrel(FILE *out, struct PATCH *patch, int pos, int rip)
   {
   const char *op;
   long disp;
   long newdisp;
   int i;
   int sz;
   boolean err;
   char sgn;

   err = FALSE;
   switch (rip)
      {
      case 1 :
                /* jmp */
         if ((patch->data[pos] & 255) == 0xeb) op = "jmp";
         else op = condjmp(patch->data[pos] & 15);
         disp = patch->data[pos + 1];
         if (disp & 0x80) disp |= -128;
         break;
      case 4 :
              /* near calls and jumps only */
         disp = patch->data[pos + rip];
         if (disp & 128) disp |= -128;
         for (i=1; i<4; i++)
            disp = (disp << 8) + (patch->data[pos + rip - i] & 255);
         if (rip == 4) op = (patch->data[pos] & 1 ? "jmp" : "call");
         else op = condjmp(patch->data[pos + 1] & 15);
         break;
      default :
// TODO immediate data after relative address
               /* something [rip], or conditional jumps */
         sz = instrsize(&patch->data[pos],0);
         if ((rip + 1) != sz)
            {
            fprintf(stderr,"Unsupported relative + immediate\n");
            err = TRUE;
            }
         disp = patch->data[pos + rip];
         if (disp & 128) disp |= -128;
         for (i=1; i<4; i++)
            disp = (disp << 8) + (patch->data[pos + rip - i] & 255);
         op = (const char*)NULL;
         break;
      }
   newdisp = pos + rip + 1 + disp - 5;
   if (newdisp & 0x80000000)
      {
      newdisp = -newdisp;
      sgn = '-';
      }
   else
      sgn = '+';
   if (op)
      {
#if INTELSYNTAX
      if (patch->type == TTRACE)
         fprintf(out, " %s near %s%c%ld\n", op, patch->back, sgn, newdisp);
      else
         fprintf(out, " %s near back%s%c%ld\n", op, patch->name, sgn, newdisp);
#else
      if (patch->type == TTRACE)
         fprintf(out," %s %s+c%ld\n", op, patch->back, sgn, newdisp);
      else
         fprintf(out," %s back%s%c%lx\n", op, patch->name, sgn, newdisp);
#endif
      }
   else
      {
#if INTELSYNTAX
      fprintf(out," db 0%02xH", patch->data[pos] & 255);
      for (i=1; i<rip-3; i++)
         fprintf(out,", 0%02xH", patch->data[pos + i] & 255);
      fprintf(out,"\n");
      if (patch->type == TTRACE)
         fprintf(out," dd %s-*%c%ld\n",patch->back, sgn, newdisp);
      else
         fprintf(out," dd back%s-*%c%ld\n",patch->name, sgn, newdisp);
#else
      fprintf(out," .byte %d", patch->data[pos] & 255);
      for (i=1; i<rip-3; i++)
         fprintf(out,", %d", patch->data[pos + i] & 255);
      fprintf(out,"\n");
      if (patch->type == TTRACE)
         fprintf(out," .long %s-next%s%c%ld\n",
                          patch->name, patch->back, sgn, newdisp);
      else
         fprintf(out," .long back%s-next%s%c%ld\n",
                          patch->name, patch->name, sgn, newdisp);
      fprintf(out,"next%s :\n", patch->name);
#endif
      }
   return (err);
   }

boolean buildtraces(FILE *out)
   {
   struct PATCH *patch;
   int sz;
   int rip;
   int next;
   boolean err;
   boolean needrelay;
   int i;

   err = FALSE;
   needrelay = FALSE;
   for (patch=firstpatch; patch && !err; patch=patch->next)
      if ((patch->type == TTRACE) || (patch->type == TCALL)
        || (patch->type == TINSERT))
         {
#if INTELSYNTAX
         if (patch->type == TTRACE)
            {
            fprintf(out,"%s : ; trace at 0x%lx\n",patch->name,(long)patch->loc);
            fprintf(out," call qword relaydump[rip]\n");
            }
         else
            {
            fprintf(out,"go%s : ; inserted code at 0x%lx\n",patch->name,(long)patch->loc);
            if (patch->type == TCALL)
               {
               fprintf(out," extrn %s\n",patch->name);
               fprintf(out," call near %s\n",patch->name);
               }
           else
              fprintf(out,"%s\n",patch->code);
           }
#else
         if (patch->type == TTRACE)
            {
            fprintf(out,"%s :\n",patch->name);
            fprintf(out," call *relaydump(%%rip)\n");
            }
         else
            {
            fprintf(out,"go%s :\n",patch->name);
            if (patch->type == TCALL)
               {
               fprintf(out," .extern %s\n",patch->name);
               fprintf(out," call %s\n",patch->name);
               }
            else
               fprintf(out,"%s\n",patch->code);
            }
#endif
             /* fix the first instruction offset if relative */
         sz = instrsize(patch->data, 0);
         rip = relative(patch->data);
         if (rip)
            {
            if (relocrel(out, patch, 0, rip)) err = TRUE;
            next = rip + 1;
            }
         else next = 0;
         if (sz < patch->size)
            {
            rip = relative(&patch->data[sz]);
            if (rip)
               {
                   /* fix the second instruction offset if relative */
               if (!next)
                  {
                      /* copy the first instruction, unless previously done */
#if INTELSYNTAX
                  fprintf(out," db 0%02xH", patch->data[0] & 255);
                  for (i=1; i<sz; i++)
                     fprintf(out,", 0%02xH", patch->data[i] & 255);
                  fprintf(out,"\n");
#else
                  fprintf(out," .byte %d", patch->data[0] & 255);
                  for (i=1; i<sz; i++)
                     fprintf(out,", %d", patch->data[i] & 255);
                  fprintf(out,"\n");
#endif
                  next = sz;
                  }
               if (relocrel(out, patch, sz, rip))
                  err = TRUE;
               next += rip + 1;
               }
            }
         if (next < patch->size)
            {
#if INTELSYNTAX
            fprintf(out," db 0%02xH", patch->data[next] & 255);
            for (i=next+1; i<patch->size; i++)
               fprintf(out,", 0%02xH", patch->data[i] & 255);
            fprintf(out,"\n");
#else
            fprintf(out," .byte %d", patch->data[next] & 255);
            for (i=next+1; i<patch->size; i++)
               fprintf(out,", %d", patch->data[i] & 255);
            fprintf(out,"\n");
#endif
            }
#if INTELSYNTAX
         if (patch->type == TTRACE)
            fprintf(out," jmp near %s\n",patch->back);
         else fprintf(out," jmp near back%s\n",patch->name);
#else
         if (patch->type == TTRACE)
            fprintf(out," jmp %s\n",patch->back);
         else fprintf(out," jmp back%s\n",patch->name);
#endif
         needrelay = (patch->type == TTRACE);
         }
   if (needrelay)
      {
#if INTELSYNTAX
      fprintf(out,"          align 16\n");
      fprintf(out,"relaydump dq dumpregs\n");
      fprintf(out,"          dq drv_data\n");
#else
      fprintf(out,"          .align 16\n");
      fprintf(out,"relaydump :\n");
      fprintf(out,"    .quad dumpregs\n");
      fprintf(out,"    .quad drv_data\n");
#endif
      }
   return (err);
   }

/*
 *            Patch a single movdqu/mpvaps
 */

boolean patchxmm(FILE *out, struct PATCH *patch)
   {
   REGS tmpreg;
   char dstname[16];
   char srcname[16];
   int offs;
   boolean err;

   err = FALSE;
   if ((patch->dstreg >= XMM0) && (patch->dstreg <= XMM7))
      {
      if ((patch->srcreg != RAX) && (patch->xreg != RAX))
         tmpreg = RAX;
      else
         if ((patch->srcreg != RDX) && (patch->xreg != RDX))
            tmpreg = RDX;
         else
            tmpreg = RCX;
      if (patch->srcreg == RIP)
         {
#if INTELSYNTAX
         fprintf(out," push rdx\n");
         fprintf(out," lea rdx,%s[rip]\n",patch->back);
#else
         fprintf(out," pushq %%rdx\n");
         fprintf(out," leaq %s(%%rip),%%rdx\n",patch->back);
#endif
         strcpy(srcname, "rdx");
         offs = patch->size - 5;
         }
      else
         {
         if (patch->xreg == RSP)
            strcpy(srcname, regs[patch->srcreg]);
         else
            if (patch->xshift == 1)
               {
#if INTELSYNTAX
               sprintf(srcname,"%s+%s",
                    regs[patch->srcreg],regs[patch->xreg]);
#else
               sprintf(srcname,"%s,%%%s",
                    regs[patch->srcreg],regs[patch->xreg]);
#endif
               }
            else
               {
#if INTELSYNTAX
               sprintf(srcname,"%s+%s*%d",
                    regs[patch->srcreg],regs[patch->xreg],patch->xshift);
#else
               sprintf(srcname,"%s,%%%s,%d",
                    regs[patch->srcreg],regs[patch->xreg],patch->xshift);
#endif
               }
         offs = (patch->srcreg == RSP ? 8 : 0);
         }
                 /* Modifying the condition code is not allowed */
#if INTELSYNTAX
      fprintf(out," push %s\n",regs[tmpreg]);
      fprintf(out," mov %s,%d+%d[%s]\n",regs[tmpreg],(int)patch->offset,offs,srcname);
      fprintf(out," mov r%sl[rip],%s\n",regs[patch->dstreg],regs[tmpreg]);
      fprintf(out," mov %s,%d+%d+8[%s]\n",regs[tmpreg],(int)patch->offset,offs,srcname);
      fprintf(out," mov r%sh[rip],%s\n",regs[patch->dstreg],regs[tmpreg]);
      fprintf(out," pop %s\n",regs[tmpreg]);
      if (patch->srcreg == RIP)
         fprintf(out," pop rdx\n");
#else
      fprintf(out," pushq %%%s\n",regs[tmpreg]);
      fprintf(out," movq %d+%d(%%%s),%%%s\n",(int)patch->offset,offs,srcname,regs[tmpreg]);
      fprintf(out," movq %%%s,r%sl(%%rip)\n",regs[tmpreg],regs[patch->dstreg]);
      fprintf(out," movq %d+%d+8(%%%s),%%%s\n",(int)patch->offset,offs,srcname,regs[tmpreg]);
      fprintf(out," movq %%%s,r%sh(%%rip)\n",regs[tmpreg],regs[patch->dstreg]);
      fprintf(out," popq %%%s\n",regs[tmpreg]);
      if (patch->srcreg == RIP)
         fprintf(out," popq %%rdx\n");
#endif
      }
   else
      {
      if ((patch->dstreg != RAX) && (patch->xreg != RAX))
         tmpreg = RAX;
      else
         if ((patch->dstreg != RDX) && (patch->xreg != RDX))
            tmpreg = RDX;
         else
            tmpreg = RCX;
      if (patch->dstreg == RIP)
         {
#if INTELSYNTAX
         fprintf(out," push rdx\n");
         fprintf(out," lea rdx,%s[rip]\n",patch->back);
#else
         fprintf(out," pushq %%rdx\n");
         fprintf(out," leaq %s(%%rip),%%rdx\n",patch->back);
#endif
         strcpy(dstname, "rdx");
         offs = patch->size - 5;
         }
      else
         {
         if (patch->xreg == RSP)
            strcpy(dstname, regs[patch->dstreg]);
         else
            if (patch->xshift == 1)
               {
#if INTELSYNTAX
               sprintf(dstname,"%s+%s",
                    regs[patch->dstreg],regs[patch->xreg]);
#else
               sprintf(dstname,"%s,%%%s",
                    regs[patch->dstreg],regs[patch->xreg]);
#endif
               }
            else
               {
#if INTELSYNTAX
               sprintf(dstname,"%s+%s*%d",
                    regs[patch->dstreg],regs[patch->xreg],patch->xshift);
#else
               sprintf(dstname,"%s,%%%s,%d",
                    regs[patch->dstreg],regs[patch->xreg],patch->xshift);
#endif
               }
         offs = (patch->dstreg == RSP ? 8 : 0);
         }
                 /* Modifying the condition code is not allowed */
#if INTELSYNTAX
      fprintf(out," push %s\n",regs[tmpreg]);
      fprintf(out," mov %s,r%sl[rip]\n",regs[tmpreg],regs[patch->srcreg]);
      fprintf(out," mov %d+%d[%s],%s\n",(int)patch->offset,offs,dstname,regs[tmpreg]);
      fprintf(out," mov %s,r%sh[rip]\n",regs[tmpreg],regs[patch->srcreg]);
      fprintf(out," mov %d+%d+8[%s],%s\n",(int)patch->offset,offs,dstname,regs[tmpreg]);
      fprintf(out," pop %s\n",regs[tmpreg]);
      if (patch->srcreg == RIP)
         fprintf(out," pop rdx\n");
#else
      fprintf(out," pushq %%%s\n",regs[tmpreg]);
      fprintf(out," movq r%sl(%%rip),%%%s\n",regs[patch->srcreg],regs[tmpreg]);
      fprintf(out," movq %%%s,%d+%d(%%%s)\n",regs[tmpreg],(int)patch->offset,offs,dstname);
      fprintf(out," movq r%sh(%%rip),%%%s\n",regs[patch->srcreg],regs[tmpreg]);
      fprintf(out," movq %%%s,%d+%d+8(%%%s)\n",regs[tmpreg],(int)patch->offset,offs,dstname);
      fprintf(out," popq %%%s\n",regs[tmpreg]);
      if (patch->srcreg == RIP)
         fprintf(out," popq %%rdx\n");
#endif
      }
   return (err);
   }

boolean buildpatches(FILE *out)
   {
   struct PATCH *patch;
   boolean err;
   int offset;
#if INTELSYNTAX
#else
   int nextnum;
#endif
   int rip;
   int i;

   err = FALSE;
#if INTELSYNTAX
   fprintf(out,"          align 16\n");
#else
   nextnum = 0;
   fprintf(out,"          .align 16\n");
#endif
   for (patch=firstpatch; patch && !err; patch=patch->next)
      if (patch->type == TXMM)
         {
         if (patch->next && (patch->next->flags & PDOUBLE))
            {
#if INTELSYNTAX
            fprintf(out,"%s : ; inserting for 0x%lx\n",
                            patch->next->name,(long)patch->next->loc);
            err = patchxmm(out, patch->next) || patchxmm(out, patch);
            fprintf(out," jmp near %s\n",patch->next->back);
#else
            fprintf(out,"%s :\n",patch->next->name);
            err = patchxmm(out, patch->next) || patchxmm(out, patch);
            fprintf(out," jmp %s\n",patch->next->back);
#endif
            patch = patch->next;
            }
         else
            {
#if INTELSYNTAX
            fprintf(out,"%s : ; inserting at 0x%lx\n",patch->name,(long)patch->loc);
#else
            fprintf(out,"%s :\n",patch->name);
#endif
            err = patchxmm(out, patch);
            if (patch->flags & PNEXT)
               {
                 /*
                  * We had to relocate two instructions, and we assume
                  * there is no jmp to the second one.
                  */
               rip = relative(patch->data);
               if (rip)
                  {
                     /* beginning of instruction */
#if INTELSYNTAX
                  fprintf(out," db 0%02xH",patch->data[0] & 255);
                  for (i=1; i<rip-3; i++)
                     fprintf(out,", 0%02xH",patch->data[i] & 255);
#else
                  fprintf(out," .byte %d",patch->data[0] & 255);
                  for (i=1; i<rip-3; i++)
                     fprintf(out,", %d",patch->data[i] & 255);
#endif
                  fprintf(out,"\n");
                     /* compute offset, which is relative to old ending */
                  offset = ((((((patch->data[rip] & 255) << 8)
                              + (patch->data[rip-1] & 255)) << 8)
                              + (patch->data[rip-2] & 255)) << 8)
                              + (patch->data[rip-3] & 255);
                    /* make offset relative to back location */
                  offset += patch->size + patch->nextsize - 5;
#if INTELSYNTAX
                  if (offset < 0)
                     fprintf(out," dd %s-*%d\n",patch->back,(int)offset);
                  else
                     fprintf(out," dd %s-*+%d\n",patch->back,(int)offset);
#else
                  nextnum++;
                  if (offset < 0)
                     fprintf(out," .long %s-xmmnext%d%d\n",
                                     patch->back,nextnum,(int)offset);
                  else
                     fprintf(out," .long %s-xmmnext%d+%d\n",
                                     patch->back,nextnum,(int)offset);
                  fprintf(out,"xmmnext%d :\n",nextnum);
#endif
                     /* there could be an immediate value here */
                  if ((rip + 1) != patch->nextsize) err = TRUE;
                  }
               else
                  {
#if INTELSYNTAX
                  fprintf(out," db 0%02xH",patch->data[0] & 255);
                  for (i=1; i<patch->nextsize; i++)
                     fprintf(out,", 0%02xH",patch->data[i] & 255);
#else
                  fprintf(out," .byte %d",patch->data[0] & 255);
                  for (i=1; i<patch->nextsize; i++)
                     fprintf(out,", %d",patch->data[i] & 255);
#endif
                  fprintf(out,"\n");
                  }
               }
#if INTELSYNTAX
            fprintf(out," jmp near %s\n",patch->back);
#else
            fprintf(out," jmp %s\n",patch->back);
#endif
            }
         }
   return (err);
   }

boolean buildcode(const char *fileout)
   {
   const char *ptr;
   FILE *out;
   long long v;
   long loc;
   int ptrsz;
   int ptrbits;
   const char *ptrdir;
   const char *imp;
   struct PATCH *patch;
   const char *p;
   char *ahead;
   int aheadcnt;
   boolean err;
   int immsz;
   int skip;
   int size;
   int rip;
   int sz;
   int c;
   int i;
   int m;
   int cnt;
#if INTELSYNTAX
#else
   int nextnum;
#endif

   err = FALSE;
   if (fileout)
      {
      out = fopen(fileout,"w");
      if (!out)
         {
         fprintf(stderr,"Could not create %s\n",fileout);
         err = TRUE;
         }
      }
   else out = stdout;
   if (!err && !buildheader(out))
      {
      ptrsz = (x8664 ? 8 : 4);
      ptrbits = 8*ptrsz;
#if INTELSYNTAX
      fprintf(out,"_TEXT     segment page public '.text'\n");
      fprintf(out,"__TEXT    proc  near\n");
      ptrdir = (ptrsz == 8 ? "dq" : "dd");
#else
      nextnum = 0;
      fprintf(out,"   .text\n");
      ptrdir = (ptrsz == 8 ? ".quad" : ".int");
#endif
      fprintf(out,"drv_data :\n");
      cnt = 0;
      aheadcnt = 0;
      patchoffset = 8;
      ptr = img;
      for (loc=0; (loc<filesize) && !err; loc++)
         {
         c = (--aheadcnt < 0 ? *ptr++ : *ahead++) & 255;
         m = (relocs[loc/4] >> ((loc & 3)*2)) & 3;
         switch (m)
            {
            case RNONE :
#if INTELSYNTAX
               if (cnt == 0) fprintf(out," db ");
               else fprintf(out,",");
               fprintf(out," 0%02xH",c);
#else
               if (cnt == 0) fprintf(out," .byte ");
               else fprintf(out,",");
               fprintf(out," %d",c);
#endif
               cnt++;
               if ((cnt >= 8) || !((loc + 1) & 7))
                  {
#if INTELSYNTAX
                  fprintf(out," ; %06lx\n",loc - cnt + 1);
#else
                  fprintf(out,"\n");
#endif
                  cnt = 0;
                  }
               break;
            case RRELOC :
               v = c;
               for (i=8; i<ptrbits; i+=8)
                  v += ((long long)
                          ((--aheadcnt < 0 ? *ptr++ : *ahead++) & 255)) << i;
               if (cnt > 0) fprintf(out,"\n");
               if ((v >= (long long)rvavars) && (v < (long long)(rvavars + sizevars)))
                  {
fprintf(stderr,"At loc 0x%lx ptr to var : RVA 0x%lx (0x%lx 0x%lx)\n",(long)loc,(long)v,(long)rvavars,(long)(rvavars + sizevars));
#if INTELSYNTAX
                  fprintf(out," %s variables+0%llxH ; %06lx\n",
                                       ptrdir, v - rvavars, loc);
#else
                  fprintf(out," %s variables+%lld\n",
                                       ptrdir, v - rvavars);
#endif
                  }
               else
                  {
#if INTELSYNTAX
                  fprintf(out," %s drv_data+0%llxH ; %06lx\n",
                                       ptrdir, v + loccode - rvacode, loc);
#else
                  fprintf(out," %s drv_data+%lld\n",
                                       ptrdir, v + loccode - rvacode);
#endif
                  }
               loc += ptrsz - 1;
               cnt = 0;
               break;
            case RPATCH :
               patch = findpatch(loc);
               if (patch)
                  switch (patch->type)
                     {
                     case TXMM :
                        size = patch->size;
                        if (patch->flags & (PDOUBLE | PNEXT))
                           size += patch->nextsize;
#if INTELSYNTAX
                        if (cnt > 0) fprintf(out," ; 0x%lx\n",loc - cnt);
                        fprintf(out,"  jmp near %s\n",patch->name);
                        fprintf(out,"%s :\n",patch->back);
#else
                        if (cnt > 0) fprintf(out,"\n");
                        fprintf(out,"  jmp %s\n",patch->name);
                        fprintf(out,"%s :\n",patch->back);
#endif
                        skip = 5;
                        if (size > 5)
                           {
                           if (size == 6)
                              {
#if INTELSYNTAX
                              fprintf(out,"  db 090H\n");
#else
                              fprintf(out,"  .byte 144\n");
#endif
                              skip = 6;
                              }
                           else
                              {
#if INTELSYNTAX
                              fprintf(out,"  db 0EBH, 0%xH\n",size - 7);
#else
                              fprintf(out,"  .byte 235, %d\n",size - 7);
#endif
                              if (size > 7)
                                 {
#if INTELSYNTAX
                                 fprintf(out,"  db 090H");
                                 for (i=8; i<size; i++)
                                    fprintf(out,", 090H");
#else
                                 fprintf(out,"  .byte 144");
                                 for (i=8; i<size; i++)
                                    fprintf(out,", 144");
#endif
                                 fprintf(out,"\n");
                                 }
                              skip = size;
                              }
                           }
                        for (i=1; i<skip; i++)
                            (--aheadcnt < 0 ? *ptr++ : *ahead++);
                        loc += skip - 1;
                        cnt = 0;
                        break;
                     case TPATCHSYST :
                        v = c;
                        for (i=8; i<ptrbits; i+=8)
                           v += ((long long)
                              ((--aheadcnt < 0 ? *ptr++ : *ahead++) & 255)) << i;
#if INTELSYNTAX
                        if (cnt > 0) fprintf(out," ; 0x%lx\n",loc - cnt);
                        fprintf(out," %s %s ; %06lx\n",ptrdir,patch->name,loc);
#else
                        if (cnt > 0) fprintf(out,"\n");
                        fprintf(out," %s %s\n",ptrdir,patch->name);
#endif
                        loc += ptrsz - 1;
                        cnt = 0;
                        break;
                     case TPATCHR :
// obsolete
                        if (!strncmp(patch->name,"or dword",8)
                          || !strncmp(patch->name,"and dword",9))
                           immsz = 1;
                        else immsz = 0;
                        for (i=4; i<(patch->size - immsz); i++)
                           c = (--aheadcnt < 0 ? *ptr++ : *ahead++);
                        v = c;
                        for (i=8; i<32; i+=8)
                           v += ((long long)
                              ((--aheadcnt < 0 ? *ptr++ : *ahead++) & 255)) << i;
                        for (i=0; i<immsz; i++)
                           c = (--aheadcnt < 0 ? *ptr++ : *ahead++);
//fprintf(stderr,"old offset 0x%llx old loc 0x%lx\n",(long long)v,(long)(loc + v + patch->size));
                        if (cnt > 0) fprintf(out," ; 0x%lx\n",loc - cnt);
                        fprintf(out," ");
                        for (p=patch->name; *p; p++)
                           if (*p == '@')
                              fprintf(out,"variables+0%lXH",(long)(loc + v + patch->size - locvars));
                           else fprintf(out,"%c",*p);
                        fprintf(out," ; %06lx\n",loc);
                        loc += patch->size - 1;
                        cnt = 0;
                        break;
                     case TRIP :
                        v = c;
                        for (i=8; i<32; i+=8)
                           v += ((long long)
                              ((--aheadcnt < 0 ? *ptr++ : *ahead++) & 255)) << i;
#if INTELSYNTAX
                        if (cnt > 0) fprintf(out," ; 0x%lx\n",loc - cnt);
                        fprintf(out," dd variables-*+%ld",(long)(v + loc +4 - locvars));
                        fprintf(out," ; %06lx\n",loc);
#else
                        if (cnt > 0) fprintf(out,"\n");
                        nextnum++;
                        fprintf(out," .long variables-next%d+%ld\n",
                                    nextnum,(long)(v + loc +4 - locvars));
                        fprintf(out,"next%d :\n",nextnum);
#endif
                        loc += patch->size - 1;
                        cnt = 0;
                        break;
                     case TVAR :
                        v = c;
                        for (i=8; i<32; i+=8)
                           v += ((long long)
                              ((--aheadcnt < 0 ? *ptr++ : *ahead++) & 255)) << i;
#if INTELSYNTAX
                        if (cnt > 0) fprintf(out," ; 0x%lx\n",loc - cnt);
                        fprintf(out," dd variables+%ld",(long)(v - rvavars));
                        fprintf(out," ; %06lx\n",loc);
#else
                        if (cnt > 0) fprintf(out,"\n");
                        nextnum++;
                        fprintf(out," .long variables+%ld\n",
                                    (long)(v - rvavars));
#endif
                        loc += patch->size - 1;
                        cnt = 0;
                        break;
                     case TIMPORT :
                        imp = patch->name;
                        v = c;
                        for (i=8; i<ptrbits; i+=8)
                           v += ((long long)
                                    ((--aheadcnt < 0 ? *ptr++ : *ahead++) & 255)) << i;
#if INTELSYNTAX
                        if (cnt > 0) fprintf(out," ; 0x%lx\n",loc - cnt);
                        if (imp)
                           fprintf(out," %s %s ; %06lx\n",ptrdir,imp,loc);
                        else
                           fprintf(out," %s import+0%llxH ; %06lx\n",
                                             ptrdir,v + loccode - rvacode,loc);
#else
                        if (cnt > 0) fprintf(out,"\n");
                        if (imp)
                           fprintf(out," %s %s\n",ptrdir,imp);
                        else
                           fprintf(out," %s import+%lld\n",
                                             ptrdir, v + loccode - rvacode);
#endif
                        loc += ptrsz - 1;
                        cnt = 0;
                        break;
                     case TTRACE :
                     case TCALL :
                     case TINSERT :
                        i = 0;
                        patch->data[i++] = c;
                        while (i < 32)
                           patch->data[i++]
                              = (--aheadcnt < 0 ? *ptr++ : *ahead++);
                        size = 0;
                        do {
                           sz = instrsize(&patch->data[size],0);
                           size += sz;
                           } while (sz && (size < 5));
                        if (sz)
                           {
                           patch->size = size;
                           patch->offset = patchoffset;
                           rip = relative(patch->data);
                           if (rip == 1)
                              patchoffset += 6 + size + 5
                                      + ((patch->data[0] & 255) == 0xeb ? 3 : 4);
                           else patchoffset += 6 + size + 5;
#if INTELSYNTAX
                           if (cnt > 0) fprintf(out," ; 0x%lx\n",loc - cnt);
                           if (patch->type == TTRACE)
                              fprintf(out," jmp near %s\n",patch->name);
                           else fprintf(out," jmp near go%s\n",patch->name);
#else
                           if (cnt > 0) fprintf(out,"\n");
                           if (patch->type == TTRACE)
                              fprintf(out," jmp %s\n",patch->name);
                           else fprintf(out," jmp go%s\n",patch->name);
#endif
                           if (patch->type == TTRACE)
                              fprintf(out,"%s :\n",patch->back);
                           else fprintf(out,"back%s :\n",patch->name);
                           skip = 5;
                           }
                        else
                           {
                           fprintf(stderr,"Could not compute size at 0x%lx\n",loc);
                           skip = 0;
                           err = TRUE;
                           }
                        cnt = 0;
                        ahead = &patch->data[skip];
                        aheadcnt = 32 - skip;
                        loc += skip - 1;
                        if (skip && (size > 5))
                           {
                              /* hide unused bytes */
                           (--aheadcnt < 0 ? *ptr++ : *ahead++);
#if INTELSYNTAX
                           fprintf(out," db 0%02XH",0x90);
#else
                           fprintf(out," .byte %d",0x90);
#endif
                           for (i=6; i<size; i++)
                              {
                              (--aheadcnt < 0 ? *ptr++ : *ahead++);
#if INTELSYNTAX
                              fprintf(out,", 0%02XH",0x90);
#else
                              fprintf(out,", %d",0x90);
#endif
                              }
                           fprintf(out,"\n");
                           loc += size - 5;
                           }
                        break;
                     }
               break;
            }
         }
#if INTELSYNTAX
      if (cnt) fprintf(out," ; %06lx\n",loc - cnt);
      fprintf(out,"drv_data_check :\n");
      fprintf(out," dd drv_data_check-drv_data\n");
      fprintf(out," align 16\n");
#else
      if (cnt) fprintf(out,"\n");
      fprintf(out,"drv_data_check :\n");
      fprintf(out," .long drv_data_check-drv_data\n");
      fprintf(out," .align 16\n");
#endif
      if (!err)
         {
         if (buildtraces(out) || buildpatches(out)) err = TRUE;
#if INTELSYNTAX
         fprintf(out,"__TEXT    endp\n");
         fprintf(out,"_TEXT     ends\n");
         fprintf(out,"          end\n");
#endif
         }
      if (fileout)
         fclose(out);
      }
   return (err);
   }

#if ALONE

int main(int argc, char *argv[])
   {
   char *image;
   FILE *in;
   long loc;
   int c;
   boolean err;

   x8664 = sizeof(long) > 4;
   in = fopen((x8664 ? INFILE64 : INFILE32),"rb");
   if (in)
      {
      fseek(in, 0L, 2);
      filesize = ftell(in);
      fseek(in, 0L, 0);
      image = (char*)malloc(filesize);
      if (image)
         {
         if (x8664)
            {
            loccode = 0x1000;
            rvacode = 0x11000;
            sizecode = 0x095f78;
            locinit = 0x13b000;
            sizeinit = 0x8f0;
            locdata = 0x12ac00;
            locvars = 0x136800;
            rvavars = 0x146800;
            sizevars = 0x360;
            locreloc = 0x13d000;
            sizereloc = 0x0ec8;
            locimports = 0x13b064;
            }
         else
            {
            loccode = 0x0480;
            rvacode = 0x10480;
            sizecode = 0x073f7b;
            locinit = 0x10e400;
            sizeinit = 0x884;
            locdata = 0x104080;
            locvars = 0x104080;
            rvavars = 0x114080;
            sizevars = 0xa320;
            locreloc = 0x10f100;
            sizereloc = 0x2b58;
            locimports = 0x10e448;
            }

         loc = 0;
         c = getc(in);
         if (c != EOF)
            do {
               image[loc++] = c;
               c = getc(in);
               } while (c != EOF);
         if (loc == filesize)
            {
            img = image;
            firstpatch = (struct PATCH*)NULL;
            err = findrelocs()
              || findchanges()
              || findimports()
              || settrace()
              || setcalls()
              || buildcode((const char*)NULL);
            }
         else
            {
            fprintf(stderr,"Could not read image\n");
            err = TRUE;
            }
         }
      else
         {
         fprintf(stderr,"Could not allocate image\n");
         err = TRUE;
         }
      fclose(in);
      }
   if (err) exit(1);
   return (0);
   }

#else

boolean getlocations(void)
   {
   boolean err;
   image_optional_header opt_hdr;
   image_file_header file_hdr;
   image_section_header sect_hdr;
   long rvadata, sizedata;
   int sectioncnt;
   int i;
   vm_offset_t vmimg;
   unsigned short *pversion;
   static const unsigned short versionid[] =
                       { 'V', 'S', '_', 'V', 'E', 'R', 'S', 'I',
                         'O', 'N', '_', 'I', 'N', 'F', 'O' };

   err = FALSE;
   vmimg = (vm_offset_t)img;
   err = pe_get_optional_header(vmimg, &opt_hdr)
       || pe_get_file_header(vmimg, &file_hdr);
   if (!err)
      {
      if (file_hdr.ifh_machine == 0x8664)
         x8664 = TRUE;
      else
          if (file_hdr.ifh_machine == 0x14c)
            x8664 = FALSE;
         else
            {
            fprintf(stderr,"** Driver for unsupported target CPU\n");
            err = TRUE;
            }
      }
   if (!err)
      {
      if (!pe_get_section(vmimg, &sect_hdr, ".rsrc"))
         {
         static const char *months[]
             = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" } ;
         unsigned int day; /* 0 <= day <= 50816, from Jan 1st, 1901 */
         unsigned int mon;
         unsigned int cnt;

         day = file_hdr.ifh_timestamp/86400 + 25202;
         year = 1901;
                               /* periods of a four years */
         cnt = day/1461;
         day -= 1461*cnt;
         year += 4*cnt;  
                               /* periods of a single year */
         cnt = (3*day + 3)/1096;
         day -= 365*cnt;
         year += cnt;
                               
         if ((year % 4) && (day > 58)) day++;
         if (day > 59)
            {
            mon = (5*day + 161)/153;
            day -= (153*mon - 162)/5;
            }
         else
            {
            mon = day/31 + 1;
            day -= 31*(mon - 1) - 1;
            }
           /* Avoid navigating in resource directory levels... */
         pversion = (unsigned short*)&img[sect_hdr.ish_rawdataaddr];
         i = 0;
         while ((i < 512)
           && memcmp((char*)pversion,(char*)versionid, sizeof(versionid)))
            {
            i += 2;
            pversion++;
            }
         if (i < 512)
            fprintf(stderr,"Driver version %u.%u.%u.%u, %s %d, %d\n",
                  pversion[22],pversion[21],pversion[24],pversion[23],
                  months[mon - 1], day, year);
         }
      switch (file_hdr.ifh_timestamp)
         {
         case 0x46f34344 : /* Sep 21 04:06:28 2007 UTC */
            year = 2007;
            break;
         case 0x47e45cf8 : /* Mar 22 01:12:24 2008 UTC */
            year = 2008;
            break;
         case 0x4ba830b5 : /* Mar 23 03:08:37 2010 UTC */
            year = 2010;
            break;
         default :
            fprintf(stderr,"*** Warning : untested driver\n");
            year = 0;
            break;
         }
      locimports = opt_hdr.ioh_datadir[IMAGE_DIRECTORY_ENTRY_IMPORT].idd_vaddr;
      locreloc = opt_hdr.ioh_datadir[IMAGE_DIRECTORY_ENTRY_BASERELOC].idd_vaddr;
      sizereloc = opt_hdr.ioh_datadir[IMAGE_DIRECTORY_ENTRY_BASERELOC].idd_size;
      sectioncnt = pe_numsections(vmimg);
      for (i=0; (i<sectioncnt) && !err; i++)
         {
         err = pe_get_section_header(vmimg, i, &sect_hdr);
         if (!err)
            {
            if (verbose)
               fprintf(stderr,"loc %06lx rva 0x%10llx size 0x%06lx %s\n",
                      (long)sect_hdr.ish_rawdataaddr,
                      (long long)(sect_hdr.ish_vaddr + opt_hdr.ioh_imagebase),
                      (long)sect_hdr.ish_rawdatasize,
                      sect_hdr.ish_name); 
            if (!strcmp(sect_hdr.ish_name, ".text"))
               {
               loccode = sect_hdr.ish_rawdataaddr;
               sizecode = sect_hdr.ish_rawdatasize;
               rvacode = sect_hdr.ish_vaddr + opt_hdr.ioh_imagebase;
               if (verbose)
                  fprintf(stderr,"loccode 0x%lx sizecode 0x%lx rvacode 0x%lx\n",
                            (long)loccode,(long)sizecode,(long)rvacode);
               }
            if (!strcmp(sect_hdr.ish_name, ".data"))
               {
               locdata = sect_hdr.ish_rawdataaddr;
               rvadata = sect_hdr.ish_vaddr + opt_hdr.ioh_imagebase;
               sizedata = sect_hdr.ish_rawdatasize;
               if ((long)sect_hdr.ish_misc.ish_vsize <= sizedata)
                  {
                    /* variables not apparent, assume all data are variables */
                  sizedata = 0;
                  sizevars = sect_hdr.ish_misc.ish_vsize;
fprintf(stderr,"locdata 0x%lx rvadata 0x%lx sizedata 0x%lx\n",(long)locdata,(long)rvadata,(long)sizedata);
fprintf(stderr,"sizedata 0x%lx vsize 0x%lx\n",(long)sizedata,(long)sect_hdr.ish_misc.ish_vsize);
                  }
               else
                  {
                    /* variables are apparent after initialized data */
                  sizevars = sect_hdr.ish_misc.ish_vsize - sizedata;
                  }
               locvars = locdata + sizedata;
               rvavars = rvadata + sizedata;
               if (verbose)
                  {
                    /*
                     * We assume that the initialized part of .data
                     * only contains constants, and that variables
                     * are grouped in the uninitialized part.
                     */
                  fprintf(stderr,"locvars 0x%lx rvavars 0x%lx sizevars 0x%lx\n",
                            (long)locvars,(long)rvavars,(long)sizevars);
                  }
               }
            }
         }
      }
   return (err);
   }

int ndislink(const char *image, long size, const char *outfile)
   {
   boolean err;

   img = image;
   filesize = size;
   firstpatch = (struct PATCH*)NULL;
   err = getlocations()
      || findrelocs()
      || findchanges()
      || findimports()
      || settrace()
      || setcalls()
      || buildcode(outfile);
   return (err);
   }

#endif
